import argparse
import importlib.metadata

import argcomplete
from rich_argparse import RichHelpFormatter

from bipackage.autocomplete import _tab_autocomplete
from bipackage.constants import PARSED_GTF_PATH_GRCh38, WHOLE_GENE_LOCS_PATH_GRCh38
from bipackage.src import (
    bam_counts,
    bedfilegenerator,
    check_gzip_validity,
    check_reconnect,
    compile_bam_stats,
    downsample,
    fastq_read_counter,
    fastqvalidate,
    is_mounted,
    md5sumchecker,
    merge_it,
    mount_server,
    nipt_bcl2fastq,
    panelgenequery,
    remove_undetermined_fastq,
    undetermined_demultiplexer,
)
from bipackage.util._colors import blue, bold

_tab_autocomplete()

RichHelpFormatter.styles["argparse.args"] = "bold dark_cyan"
RichHelpFormatter.styles["argparse.groups"] = "bold dark_orange"
RichHelpFormatter.styles["argparse.help"] = "bold grey82"
RichHelpFormatter.styles["argparse.metavar"] = "grey35"
RichHelpFormatter.styles["argparse.syntax"] = "bold bold"
RichHelpFormatter.styles["argparse.text"] = "bold default"
RichHelpFormatter.styles["argparse.prog"] = "bold grey50"
RichHelpFormatter.styles["argparse.default"] = "italic"


def main():
    parser = argparse.ArgumentParser(description="BIpackage CLI", formatter_class=RichHelpFormatter)
    # --version
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=bold(f"BIpackage {blue(importlib.metadata.version('bipackage'))}"),
        help="Show version",
    )

    # init subparsers
    subparsers = parser.add_subparsers(dest="command")

    # ======================================== BAMTOOLS ========================================

    # ---------------------------------------- bam_counts ----------------------------------------
    bam_counts_subparser = subparsers.add_parser(
        "bam_counts", help="Get counts from BAM file.", formatter_class=RichHelpFormatter
    )
    bam_counts_subparser.add_argument("--dir", "-d", required=True, help="Input directory.")
    bam_counts_subparser.add_argument("--exome-bait", "-e", required=True, help="Exome bait.")
    bam_counts_subparser.add_argument("--num-threads", "-n", type=int, default=40, help="Number of threads to use.")

    # ------------------------------------- compile_bam_stats -------------------------------------
    compile_bam_stats_subparser = subparsers.add_parser(
        "compile_bam_stats", aliases=["cbs"], help="Complies BAM stats to CSV.", formatter_class=RichHelpFormatter
    )
    compile_bam_stats_subparser.add_argument("--dir", "-d", required=True, help="Root directory for the operations.")
    compile_bam_stats_subparser.add_argument("--output-csv", "-o", required=True, help="Output csv file.")

    # ======================================== BEDTOOLS ========================================

    # ------------------------------------- bedfilegenerator -------------------------------------
    bedfilegenerator_subparser = subparsers.add_parser(
        "bedfilegenerator",
        aliases=["bfg"],
        help="Generate and sort BED files from a parsed GTF file.",
        formatter_class=RichHelpFormatter,
    )
    bedfilegenerator_subparser.add_argument(
        "--gene-list", "-g", nargs="+", required=True, help="Gene names to include."
    )
    bedfilegenerator_subparser.add_argument("--bed-file-name", "-b", required=True, help="Name of the output BED file.")
    bedfilegenerator_subparser.add_argument(
        "--output-folder", "-o", required=True, help="Folder to save the output BED file."
    )
    bedfilegenerator_subparser.add_argument(
        "--whole-gene-list", "-w", nargs="+", required=True, help="Whole gene names to include."
    )
    bedfilegenerator_subparser.add_argument(
        "--parsed-gtf-path", "-p", default=PARSED_GTF_PATH_GRCh38, help="Path to Parsed GTF file."
    )
    bedfilegenerator_subparser.add_argument(
        "--whole-gene-locs-path", "-wglp", default=WHOLE_GENE_LOCS_PATH_GRCh38, help="Path to whole gene locs file."
    )
    bedfilegenerator_subparser.add_argument(
        "--cds", "-c", action="store_true", help="If used, use CDS features; otherwise use exon features."
    )
    bedfilegenerator_subparser.add_argument("--ref-folder", "-rf", help="Path to the folder for reference files.")

    panelgenequery_subparser = subparsers.add_parser(
        "panelgenequery",
        aliases=["pgq"],
        help="Using an intersected bedfile, produces a file with gene presence information.",
        formatter_class=RichHelpFormatter,
    )
    panelgenequery_subparser.add_argument("--bedfile", "-b", required=True, help="Path to the intersected bedfile.")
    panelgenequery_subparser.add_argument("--genes-list-file", "-g", required=True, help="Path to the gene list file.")
    panelgenequery_subparser.add_argument(
        "--gene-column-index",
        "-i",
        type=int,
        default=-1,
        help="Index of the column in the bedfile that contains the genes.",
    )
    # ======================================== FASTQTOOLS ========================================

    # ------------------------------------- downsample -------------------------------------
    downsample_subparser = subparsers.add_parser(
        "downsample",
        help="Pipeline to map, deduplicate, and downsample sequencing reads.",
        formatter_class=RichHelpFormatter,
    )
    downsample_subparser.add_argument("--sample_id", "-s", required=True, help="Sample ID.")
    downsample_subparser.add_argument("--r1", "-r1", required=True, help="Path to R1 fastq file.")
    downsample_subparser.add_argument("--r2", "-r2", required=True, help="Path to R2 fastq file.")
    downsample_subparser.add_argument("--out-path", "-o", required=True, help="Output path for the results.")
    downsample_subparser.add_argument("--reference", "-r", required=True, help="Path to the reference genome.")
    downsample_subparser.add_argument("--threads", "-t", type=int, default=40, help="Number of threads to use.")
    downsample_subparser.add_argument("--remove-all-dups", "-ra", action="store_true", help="Remove all duplicates")
    downsample_subparser.add_argument(
        "--remove_seq_dups", "-rs", action="store_true", help="Remove sequencing duplicates"
    )
    downsample_subparser.add_argument("--use-gatk-md", "-ug", action="store_true", help="Use GATK MarkDuplicatesSpark")
    downsample_subparser.add_argument("--strategy", "-st", default="HighAccuracy", help="Downsampling strategy")
    downsample_subparser.add_argument(
        "--keep", "-k", type=float, default=0.5, help="How much read to keep? Give a ratio"
    )

    # ------------------------------------ fastq_read_counter ------------------------------------
    fastq_read_counter_subparser = subparsers.add_parser(
        "fastq_read_counter", aliases=["frc"], help="Count reads from FASTQ file.", formatter_class=RichHelpFormatter
    )
    fastq_read_counter_subparser.add_argument("--directory", "-d", required=True, help="Path to directory.")
    fastq_read_counter_subparser.add_argument("--output_path", "-o", required=True, help="Path to save csv file.")

    # ------------------------------------- fastqvalidate -------------------------------------
    fastqvalidate_subparser = subparsers.add_parser(
        "fastqvalidate",
        aliases=["fqv"],
        help="Validate fastq files in a given directory using`fastQValidator`.",
        formatter_class=RichHelpFormatter,
    )
    fastqvalidate_subparser.add_argument(
        "--dir",
        "-d",
        required=True,
        help="Directory to perform the validation of fasq files.",
    )

    # ------------------------------------- merge_it -------------------------------------
    merge_it_subparser = subparsers.add_parser(
        "merge_it", aliases=["mff"], help="Merge FASTQ files.", formatter_class=RichHelpFormatter
    )
    merge_it_subparser.add_argument("--folder-paths", "-f", nargs="+", required=True, help="List of paths to folders.")
    merge_it_subparser.add_argument("--sample-names", "-s", nargs="+", required=True, help="List of name of the files.")
    merge_it_subparser.add_argument("--output-path", "-o", required=True, help="Path to the output directory.")

    # -------------------------------- undetermined_demultiplexer --------------------------------
    undetermined_demultiplexer_subparser = subparsers.add_parser(
        "undetermined_demultiplexer",
        aliases=["ud"],
        help="Filter undetermined FASTQ files for multiple samples using index information.",
        formatter_class=RichHelpFormatter,
    )
    undetermined_demultiplexer_subparser.add_argument(
        "--sample_sheet", "-s", required=True, help="Path to the sample sheet CSV file."
    )
    undetermined_demultiplexer_subparser.add_argument(
        "--input_r1", "-r1", required=True, help="Path to the undetermined R1 FASTQ.gz file."
    )
    undetermined_demultiplexer_subparser.add_argument(
        "--input_r2", "-r2", required=True, help="Path to the undetermined R2 FASTQ.gz file."
    )
    undetermined_demultiplexer_subparser.add_argument(
        "--output_dir", "-o", required=True, help="Directory to store the filtered FASTQ files."
    )
    undetermined_demultiplexer_subparser.add_argument(
        "--json_output", "-j", required=True, help="Path to output JSON file for sample target indices."
    )
    undetermined_demultiplexer_subparser.add_argument(
        "--threads", "-t", type=int, default=4, help="Number of threads to use (default: 4)."
    )

    # ----------------------------------------- remove_undetermined_fastq -------------------------
    remove_undetermined_fastq_subparser = subparsers.add_parser(
        "remove_undetermined_fastq",
        aliases=["ruf"],
        help="Remove undetermined FASTQ files (recursively) from a directory.",
        formatter_class=RichHelpFormatter,
    )
    remove_undetermined_fastq_subparser.add_argument(
        "--folder_path", "-f", required=True, help="Path to the folder containing FASTQ files."
    )
    remove_undetermined_fastq_subparser.add_argument(
        "--substring",
        "-s",
        default="Undetermined",
        help="Substring to search for in filenames (default: 'Undetermined').",
    )
    remove_undetermined_fastq_subparser.add_argument(
        "--non-recursive",
        "-nr",
        action="store_false",
        help="If set, search for undetermined FASTQ files non-recursively.",
    )

    # ======================================== ITTOOLS ========================================

    # ------------------------------------- ismount -------------------------------------
    # subcommand 1
    ismounted_subparser = subparsers.add_parser(
        "is_mounted", help="Check if a given path is a mounted server.", formatter_class=RichHelpFormatter
    )
    ismounted_subparser.add_argument("--path", "-p", required=True, help="Path to check.")
    # subcommand 2
    mount_server_subparser = subparsers.add_parser(
        "mount_server", help="Mount a server.", formatter_class=RichHelpFormatter
    )
    mount_server_subparser.add_argument("--username", "-u", required=True, help="Username.")
    mount_server_subparser.add_argument("--server_address", "-s", required=True, help="Server address.")
    mount_server_subparser.add_argument("--mount-folder", "-m", required=True, help="Mount folder.")
    mount_server_subparser.add_argument("--password", "-p", required=True, help="Password.")
    mount_server_subparser.add_argument("--version", "-v", default=None, help="Version.")
    # subcommand 3
    check_reconnect_subparser = subparsers.add_parser(
        "check_reconnect", help="Check reconnects.", formatter_class=RichHelpFormatter
    )
    # check_reconnect_subparser.add_argument("--base-mnt", "-b", required=True, help="Base mount.")
    check_reconnect_subparser.add_argument("--config-file", "-c", required=True, help="Config file.")

    # ------------------------------------- md5sumchecker -------------------------------------
    md5sumchecker_subparser = subparsers.add_parser(
        "md5sumchecker", aliases=["md5sc"], help="Check md5sum of a file.", formatter_class=RichHelpFormatter
    )
    md5sumchecker_subparser.add_argument("--directory", "-d", required=True, help="Input directory path.")
    md5sumchecker_subparser.add_argument("--extension", "-e", required=True, help="File extension to search for.")
    md5sumchecker_subparser.add_argument(
        "--num_processes", "-n", type=int, default=2, help="Number of processes to use (default: 2)"
    )

    # -------------------------------- truncatedfchecker_single --------------------------------
    check_gzip_validity_subparser = subparsers.add_parser(
        "check_gzip_validity",
        aliases=["cgv"],
        help="Check a compressed file validity.",
        formatter_class=RichHelpFormatter,
    )
    check_gzip_validity_subparser.add_argument("file_path", help="Path to the compressed file.")

    # ======================================== NIPTTOOLS ========================================
    nipt_bcl2fastq_subparser = subparsers.add_parser(
        "nipt_bcl2fastq",
        aliases=["nb2f"],
        help="Run bcl2fastq conversion for multiple BCL folders.",
        formatter_class=RichHelpFormatter,
    )
    nipt_bcl2fastq_subparser.add_argument("--folders", "-f", nargs="+", required=True, help="NIPT folders.")
    nipt_bcl2fastq_subparser.add_argument("--part", "-p", required=True, type=int, help="NIPT Part number.")
    nipt_bcl2fastq_subparser.add_argument(
        "--names", "-n", nargs="+", required=True, help="Fastq sample names - For example 24B3043312."
    )
    nipt_bcl2fastq_subparser.add_argument(
        "--output-folder", "-o", required=True, help="Path to the output Fastq folder."
    )
    nipt_bcl2fastq_subparser.add_argument("--num-readers", "-r", type=int, default=10, help="Number of readers.")
    nipt_bcl2fastq_subparser.add_argument("--num-writers", "-w", type=int, default=10, help="Number of writers.")
    nipt_bcl2fastq_subparser.add_argument("--num-processors", "-np", type=int, default=40, help="Number of processors.")
    nipt_bcl2fastq_subparser.add_argument("--compression-level", "-cl", type=int, default=8, help="Compression level.")

    nipt_bcl2fastq_subparser.add_argument("--bcl-path", "-b", required=True, help="Path to the BCL folder.")
    nipt_bcl2fastq_subparser.add_argument("--source-path", "-s", required=True, help="Path to the source folder.")

    # PARSE ALL ARGS ------------------------------
    argcomplete.autocomplete(parser, always_complete_options=True)
    args = parser.parse_args()

    # EVALUATE ARGS ------------------------------
    # Subcommand: list
    if args.command is None:
        print(bold(f"BIpackage {blue(importlib.metadata.version('bipackage'))}"))
        parser.print_help()  # Show help if no command is provided
        return

    # Subcommand: bam_counts
    elif args.command == "bam_counts":
        bam_counts(input_dir=args.dir, exome_bait=args.exome_bait, num_threads=args.num_threads)

    # Subcommand: compile_bam_stats
    elif args.command == "compile_bam_stats":
        compile_bam_stats(root_directory=args.dir, output_csv=args.output_csv)

    # Subcommand: bedfilegenerator
    elif args.command == "bedfilegenerator":
        bedfilegenerator(
            gene_list=args.gene_list,
            bed_file_name=args.bed_file_name,
            output_folder=args.output_folder,
            whole_gene_list=args.whole_gene_list,
            parsed_gtf_path=args.parsed_gtf_path,
            whole_gene_locs_path=args.whole_gene_locs_path,
            cds=args.cds,
            ref_folder=args.ref_folder,
        )

    elif args.command == "panelgenequery":
        panelgenequery(
            bedfile=args.bedfile,
            genes_list_file=args.genes_list_file,
            gene_column_index=args.gene_column_index,
        )

    # Subcommand: downsample
    elif args.command == "downsample":
        downsample(
            sample_id=args.sample_id,
            r1=args.r1,
            r2=args.r2,
            out_path=args.out_path,
            reference=args.reference,
            threads=args.threads,
            remove_all_dups=args.remove_all_dups,
            remove_seq_dups=args.remove_seq_dups,
            use_gatk_md=args.use_gatk_md,
            strategy=args.strategy,
            keep=args.keep,
        )

    # Subcommand: fastq_read_counter
    elif args.command == "fastq_read_counter":
        fastq_read_counter(directory=args.directory, output_path=args.output_path)

    # Subcommand: fastqvalidate
    elif args.command == "fastqvalidate":
        fastqvalidate(directory=args.dir)

    # Subcommand: merge_it
    elif args.command == "merge_it":
        merge_it(folder_paths=args.folder_paths, sample_names=args.sample_names, output_path=args.output_path)

    # Subcommand: undetermined_demultiplexer
    elif args.command == "undetermined_demultiplexer":
        undetermined_demultiplexer(
            sample_sheet=args.sample_sheet,
            input_r1=args.input_r1,
            input_r2=args.input_r2,
            output_dir=args.output_dir,
            json_output=args.json_output,
            threads=args.threads,
        )

    # Subcommand: remove_undetermined_fastq
    elif args.command == "remove_undetermined_fastq":
        remove_undetermined_fastq(
            folder_path=args.folder_path,
            substring=args.substring,
            recursive=args.non_recursive,
        )

    # Subcommand: ismounted
    elif args.command == "is_mounted":
        is_mounted(folder_path=args.path)

    # Subcommand: mount_server
    elif args.command == "mount_server":
        mount_server(
            username=args.username,
            server_address=args.server_address,
            mount_folder=args.mount_folder,
            password=args.password,
            version=args.version,
        )

    # Subcommand: check_reconnect
    elif args.command == "check_reconnect":
        check_reconnect(config_file=args.config_file)

    # Subcommand: md5sumchecker
    elif args.command == "md5sumchecker":
        md5sumchecker(input_directory=args.directory, file_extension=args.extension, num_processes=args.num_processes)

    # Subcommand: check_gzip_validity
    elif args.command == "check_gzip_validity":
        check_gzip_validity(file_path=args.file_path)

    # Subcommand: nipt_bcl2fastq
    elif args.command == "nipt_bcl2fastq":
        nipt_bcl2fastq(
            nipt_folders=args.folders,
            nipt_part=args.part,
            fastq_names=args.names,
            output_folder=args.output_folder,
            num_readers=args.num_readers,
            num_writers=args.num_writers,
            num_processors=args.num_processors,
            compression_level=args.compression_level,
            source_path=args.source_path,
            bcl_path=args.bcl_path,
        )




import os
import subprocess
from concurrent.futures import ThreadPoolExecutor

import pandas as pd

from bipackage.util.utilities import timer


def count_reads_and_bases(fastq_file):
    try:
        read_count = int(subprocess.check_output(f"zcat {fastq_file} | wc -l", shell=True, text=True).strip()) // 4

        first_read_length = (
            int(subprocess.check_output(f"zcat {fastq_file} | sed -n '2p' | wc -c", shell=True, text=True).strip()) - 1
        )
        base_count = read_count * first_read_length

        return fastq_file, read_count, base_count
    except subprocess.CalledProcessError as e:
        print(f"Error processing {fastq_file}: {e}")
        return fastq_file, 0, 0


@timer
def fastq_read_counter(directory: str, output_path: str) -> None:
    """
    Count fastq.

    Parameters
    ----------
    directory : str
        Path to directory.
    output_path : str
        Path to save csv file.

    """
    files_to_process = [
        os.path.join(directory, filename)
        for filename in os.listdir(directory)
        if filename.endswith(".fastq.gz") and not filename.startswith("Undetermined")
    ]

    results = []
    with ThreadPoolExecutor() as executor:
        futures = {executor.submit(count_reads_and_bases, file): file for file in files_to_process}

        for future in futures:
            filename, read_count, base_count = future.result()
            results.append(
                {
                    "file": os.path.basename(filename),
                    "reads": read_count,
                    "bases": base_count,
                }
            )

    df = pd.DataFrame(results)

    # print(df)

    df.to_csv(output_path, index=False)

    return


if __name__ == "__main__":
    pass



# BWA dedup config
# This is a default configuration for the BWA dedup pipeline.
bwa_dedup_config = {
    "ref_fasta": None,
    "bed_file": None,
    "at": "bwtsw",
    "out_path": None,
    "r1": None,
    "r2": None,
    "sample_name": None,
    "threads": 40,
    "remove_dups": False,
    "remove_seq_dups": False,
    "use_md": False,
}

# GATK gCNV config
# This is a default configuration for the GATK CNV pipeline.
gatk_gcnv_config = {
    "input_json": None,
    "output_dir": None,
    "exome_loc": None,
    "genome_fasta": None,
    "ref_dict": None,
    "contig_pp": None,
    "class_coherence_length": "10000",
    "cnv_coherence_length": "10000",
    "min_contig_length": "46709983",
    "p_active": "0.01",
    "p_alt": "0.000001",
    "interval_psi_scale": "0.001",
    "sample_psi_scale": "0.0001",
}

# GATK CNV PON config
# This is a default configuration for the GATK CNV PON pipeline.
cnv_pon_config = {
    "output_folder": None,
    "ref_fasta": None,
    "ref_dict": None,
    "interval_list": None,
    "pon_id": None,
    "normal_bams": None,
    "gc_correction": True,
    "bin_length": 10000,
    "blacklist_intervals": None,
    "padding": 250,
    "do_impute_zeros": True,
    "number_of_eigensamples": 20,
    "feature_query_lookahead": 1000000,
    "minimum_interval_median_percentile": 10.0,
    "maximum_zeros_in_sample_percentage": 5.0,
    "maximum_zeros_in_interval_percentage": 5.0,
    "extreme_sample_median_percentile": 2.5,
    "extreme_outlier_truncation_percentile": 0.1,
    "maximum_chunk_size": 16777216,
}

# SNV IND config
# This is a default configuration for the SNV IND pipeline.
snv_ind_config = {
    "algotype": "bwtsw",
    "index_fasta": False,
    "aligner_threads": 40,
    "remove_all_duplicates": True,
    "remove_sequencing_duplicates": False,
    "use_gatk_mark_duplicates": True,
    "use_dragen": False,
    "output_folder": None,
    "run_name": "test_run2",
    "ref_fasta": None,
    "ref_dict": None,
    "interval_list": None,
    "scatter_count": 48,
    "germline_resource": None,
    "index_image": None,
    "downsampling_stride": None,
    "panel_of_normals": None,
    "variants_for_contamination": None,
    "max_reads_per_alignment_start": None,
    "max_suspicious_reads_per_alignment_start": None,
    "max_population_af": None,
    "lrom": True,
    "interval_padding": 100,
    "tumor_samples": None,
    "normal_samples": None,
}


# CNV config
# This is a default configuration for the CNV pipeline.
cnv_config = {
    "algotype": "bwtsw",
    "index_fasta": False,
    "aligner_threads": 40,
    "remove_all_duplicates": True,
    "remove_sequencing_duplicates": False,
    "use_gatk_mark_duplicates": True,
    "use_dragen": False,
    "output_folder": None,
    "run_name": None,
    "ref_fasta": None,
    "ref_dict": None,
    "interval_list": None,
    "common_sites": None,
    "pon": None,
    "blacklist_intervals": None,
    "minimum_base_quality": 30,
    "number_of_eigensamples": None,
    "minimum_total_allele_count_case": None,
    "minimum_total_allele_count_normal": None,
    "genotyping_homozygous_log_ratio_threshold": None,
    "genotyping_base_error_rate": None,
    "maximum_number_of_segments_per_chromosome": None,
    "kernel_variance_copy_ratio": None,
    "kernel_variance_allele_fraction": None,
    "kernel_scaling_allele_fraction": None,
    "kernel_approximation_dimension": None,
    "window_size": [8, 16, 32, 64, 128, 256],
    "number_of_changepoints_penalty_factor": None,
    "minor_allele_fraction_prior_alpha": None,
    "number_of_samples_copy_ratio": None,
    "number_of_burn_in_samples_copy_ratio": None,
    "number_of_samples_allele_fraction": None,
    "number_of_burn_in_samples_allele_fraction": None,
    "smoothing_credible_interval_threshold_copy_ratio": None,
    "smoothing_credible_interval_threshold_allele_fraction": None,
    "maximum_number_of_smoothing_iterations": None,
    "number_of_smoothing_iterations_per_fit": None,
    "neutral_segment_copy_ratio_lower_bound": None,
    "neutral_segment_copy_ratio_upper_bound": None,
    "outlier_neutral_segment_copy_ratio_z_score_threshold": None,
    "calling_copy_ratio_z_score_threshold": None,
    "minimum_contig_length": None,
    "maximum_copy_ratio": None,
    "point_size_copy_ratio": None,
    "point_size_allele_fraction": None,
    "tumor_samples": None,
    "normal_samples": None,
}



PARSED_GTF_PATH_GRCh38 = "Homo_sapiens.GRCh38.112.parsed.gtf"
WHOLE_GENE_LOCS_PATH_GRCh38 = "Homo_sapiens.GRCh38.112.gtf.gene_names.sorted.cleaned.bed"
REFERENCE_GENOME_GRCh38 = "ref/Homo_sapiens.GRCh38.dna.primary_assembly.fa"
REFERERENCE_GENOME_hg38 = "ref/hg38.fa"



import argparse
import importlib.metadata
import json
import shutil
from pathlib import Path

import argcomplete
from rich import print as rprint
from rich_argparse import RichHelpFormatter

from bipackage.autocomplete import _tab_autocomplete
from bipackage.pipelines import bwa_dedup, gatk_gcnv, pon, snv_indel, somatic_CNV
from bipackage.pipelines.defaults import bwa_dedup_config, cnv_config, cnv_pon_config, gatk_gcnv_config, snv_ind_config

RichHelpFormatter.styles["argparse.args"] = "bold dodger_blue1"
RichHelpFormatter.styles["argparse.groups"] = "bold deep_pink2"
RichHelpFormatter.styles["argparse.help"] = "grey82"
RichHelpFormatter.styles["argparse.metavar"] = "orange_red1"
RichHelpFormatter.styles["argparse.prog"] = "bold grey85"
RichHelpFormatter.styles["argparse.syntax"] = "bold bright_white"
RichHelpFormatter.styles["argparse.text"] = "bold grey70"
RichHelpFormatter.styles["argparse.default"] = "italic"

_tab_autocomplete()

def _dict_print(dict_obj: dict, command: str) -> None:
    rprint(f"[green]Running `bipipe {command}` with ...[/green]")
    for key, value in dict_obj.items():
        if isinstance(value, (int, float)):
            rprint(f"[grey85]{key}[/grey85] : [dark_orange]{value}[/dark_orange]")
        elif isinstance(value, str):
            rprint(f"[grey85]{key}[/grey85] : [white]{value}[/white]")
        elif isinstance(value, bool):
            rprint(f"[grey85]{key}[/grey85] : [orchid1]{value}[/orchid1]")
        elif isinstance(value, list):
            rprint(f"[grey85]{key}[/grey85] : [cyan]{', '.join(map(str, value))}[/cyan]")
        else:
            rprint(f"[grey85]{key}[/grey85] : [grey23]{value}[/grey23]")


def _print_not_found_keys(not_found_keys: list) -> None:
    msg = (
        f"\n[bold red]REQUIRED KEYS NOT FOUND IN CONFIG FILE OR COMMAND LINE ARGUMENTS...[/bold red]\n"
        f"[bold red]PLEASE PROVIDE THE FOLLOWING KEYS IN YOUR CONFIG FILE OR COMMAND LINE ARGUMENTS:[/bold red]\n"
        f"\n([cyan]{', '.join(not_found_keys)}[/cyan])\n"
    )
    rprint(msg)


def bring_config(config: str) -> None:
    entry_dir = Path(__file__).parent.parent
    config_path = entry_dir / "configs" / f"{config}.json"
    # copy config file to current directory
    if config_path.exists():
        shutil.copy(config_path, Path(".") / f"{config}.json")
    else:
        rprint(f"[bold red]Configuration file `{config}.json` not found in {entry_dir / 'configs'}[/bold red]")
        rprint("[bold green]Available configurations:[/bold green]")
        for file in (entry_dir / "configs").glob("*.json"):
            rprint(f"[grey85]{file.stem}[/grey85]")

    return


def main():
    parser = argparse.ArgumentParser(
        description="BIPackage Pipelines Entry Point",
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"bipipe {importlib.metadata.version('bipackage')}",
    )

    subparsers = parser.add_subparsers(dest="command")

    # ------------------------------------BWA DEDUP SUBPARSER ------------------------------------
    bwadedup_subparser = subparsers.add_parser(
        "bwa_dedup",
        help="Run the bwa_dedup_sp_qc_mosdepth pipeline.",
        formatter_class=RichHelpFormatter,
    )
    bwadedup_subparser.add_argument(
        "--config",
        type=str,
        # required=True,
        help="Path to the configuration file for the bwa_dedup_sp_qc_mosdepth pipeline.",
    )
    bwadedup_subparser.add_argument(
        "--ref_fasta",
        type=str,
        help="Path to the reference FASTA file.",
    )
    bwadedup_subparser.add_argument(
        "--bed_file",
        type=str,
        help="Path to the BED file for coverage analysis.",
    )
    bwadedup_subparser.add_argument(
        "--at",
        type=str,
        default="bwtsw",
        help="Algorithm type for BWA indexing (default is 'bwtsw').",
    )
    bwadedup_subparser.add_argument(
        "--out_path",
        type=str,
        help="Output directory for results.",
    )
    bwadedup_subparser.add_argument(
        "--r1",
        type=str,
        help="Path to the first read FASTQ file.",
    )
    bwadedup_subparser.add_argument(
        "--r2",
        type=str,
        help="Path to the second read FASTQ file.",
    )
    bwadedup_subparser.add_argument(
        "--sample_name",
        type=str,
        help="Sample identifier for output files.",
    )
    bwadedup_subparser.add_argument(
        "--threads",
        type=int,
        help="Number of threads to use for processing.",
    )
    bwadedup_subparser.add_argument(
        "--remove_dups",
        action="store_true",
        help="Whether to remove all duplicates (default is False).",
    )
    bwadedup_subparser.add_argument(
        "--remove_seq_dups",
        action="store_true",
        help="Whether to remove sequencing duplicates (default is False).",
    )
    bwadedup_subparser.add_argument(
        "--use_md",
        action="store_true",
        help="Whether to use MarkDuplicatesSpark (default is False).",
    )
    # ------------------------------------ GATK GCNV SUBPARSER ------------------------------------
    gatk_gcnv_subparser = subparsers.add_parser(
        "gatk_gcnv",
        help="Run the GATK GCNV pipeline.",
        formatter_class=RichHelpFormatter,
    )
    gatk_gcnv_subparser.add_argument(
        "--config",
        type=str,
        # required=True,
        help="Path to the configuration file for the GATK GCNV pipeline.",
    )
    gatk_gcnv_subparser.add_argument(
        "--input_json",
        type=str,
        help="Paths to input bams (JSON file)",
    )
    gatk_gcnv_subparser.add_argument(
        "--output_dir",
        type=str,
        help="Path to output directory",
    )
    gatk_gcnv_subparser.add_argument(
        "--exome_loc",
        type=str,
        help="Path to exome BED file",
    )
    gatk_gcnv_subparser.add_argument(
        "--genome_fasta",
        type=str,
        help="Path to genome FASTA file",
    )
    gatk_gcnv_subparser.add_argument(
        "--ref_dict",
        type=str,
        help="Path to reference dictionary file",
    )
    gatk_gcnv_subparser.add_argument(
        "--contig_pp",
        type=str,
        help="Path to contig ploidy prior file",
    )
    gatk_gcnv_subparser.add_argument(
        "--class_coherence_length",
        type=str,
        help="Class coherence length",
    )
    gatk_gcnv_subparser.add_argument(
        "--cnv_coherence_length",
        type=str,
        help="CNV coherence length",
    )
    gatk_gcnv_subparser.add_argument(
        "--min_contig_length",
        type=str,
        help="Minimum contig length",
    )
    gatk_gcnv_subparser.add_argument(
        "--p_active",
        type=str,
        help="Prior probability of treating an interval as CNV-active",
    )
    gatk_gcnv_subparser.add_argument(
        "--p_alt",
        type=str,
        help="Total prior probability of alternative copy-number states",
    )
    gatk_gcnv_subparser.add_argument(
        "--interval_psi_scale",
        type=str,
        help="Typical scale of interval-specific unexplained variance",
    )
    gatk_gcnv_subparser.add_argument(
        "--sample_psi_scale",
        type=str,
        help="Typical scale of sample-specific correction to the unexplained variance",
    )
    # ------------------------------------ PON SUBPARSER ------------------------------------
    pon_subparser = subparsers.add_parser(
        "cnv_pon",
        help="Run the Panel of Normals (PON) pipeline.",
        formatter_class=RichHelpFormatter,
    )
    pon_subparser.add_argument(
        "--config",
        type=str,
        # required=True,
        help="Path to the configuration file for the PON pipeline.",
    )
    pon_subparser.add_argument(
        "--output_folder",
        type=str,
        help="Output folder for results.",
    )
    pon_subparser.add_argument(
        "--ref_fasta",
        type=str,
        help="Path to the reference FASTA file.",
    )
    pon_subparser.add_argument(
        "--ref_dict",
        type=str,
        help="Path to the reference dictionary file.",
    )
    pon_subparser.add_argument(
        "--interval_list",
        type=str,
        help="Path to the interval list file.",
    )
    pon_subparser.add_argument(
        "--pon_id",
        type=str,
        help="Panel of Normals ID.",
    )
    pon_subparser.add_argument(
        "--normal_bams",
        nargs="+",
        type=str,
        help="List of normal BAM files.",
    )
    pon_subparser.add_argument(
        "--gc_correction",
        action="store_true",
        help="Enable GC correction.",
    )
    pon_subparser.add_argument(
        "--bin_length",
        type=int,
        help="Bin length.",
    )
    pon_subparser.add_argument(
        "--blacklist_intervals",
        type=str,
        help="Path to blacklist intervals file.",
    )
    pon_subparser.add_argument(
        "--padding",
        type=int,
        help="Padding value.",
    )
    pon_subparser.add_argument(
        "--do_impute_zeros",
        action="store_true",
        help="Impute zeros.",
    )
    pon_subparser.add_argument(
        "--number_of_eigensamples",
        type=int,
        help="Number of eigensamples.",
    )
    pon_subparser.add_argument(
        "--feature_query_lookahead",
        type=int,
        help="Feature query lookahead.",
    )
    pon_subparser.add_argument(
        "--minimum_interval_median_percentile",
        type=float,
        help="Minimum interval median percentile.",
    )
    pon_subparser.add_argument(
        "--maximum_zeros_in_sample_percentage",
        type=float,
        help="Maximum zeros in sample percentage.",
    )
    pon_subparser.add_argument(
        "--maximum_zeros_in_interval_percentage",
        type=float,
        help="Maximum zeros in interval percentage.",
    )
    pon_subparser.add_argument(
        "--extreme_sample_median_percentile",
        type=float,
        help="Extreme sample median percentile.",
    )
    pon_subparser.add_argument(
        "--extreme_outlier_truncation_percentile",
        type=float,
        help="Extreme outlier truncation percentile.",
    )
    pon_subparser.add_argument(
        "--maximum_chunk_size",
        type=int,
        help="Maximum chunk size.",
    )
    # ------------------------------------ SNV/INDEL SUBPARSER ------------------------------------
    snv_ind_subparser = subparsers.add_parser(
        "snv_indel",
        help="Run the SNV/Indel pipeline.",
        formatter_class=RichHelpFormatter,
    )
    snv_ind_subparser.add_argument(
        "--config",
        type=str,
        # required=True,
        help="Path to the configuration file for the SNV/Indel pipeline.",
    )
    snv_ind_subparser.add_argument(
        "--algotype",
        type=str,
        help="Algorithm type for BWA indexing.",
    )
    snv_ind_subparser.add_argument(
        "--index_fasta",
        action="store_true",
        help="Whether to index the FASTA file.",
    )
    snv_ind_subparser.add_argument(
        "--aligner_threads",
        type=int,
        help="Number of threads for the aligner.",
    )
    snv_ind_subparser.add_argument(
        "--remove_all_duplicates",
        action="store_true",
        help="Remove all duplicates.",
    )
    snv_ind_subparser.add_argument(
        "--remove_sequencing_duplicates",
        action="store_true",
        help="Remove sequencing duplicates.",
    )
    snv_ind_subparser.add_argument(
        "--use_gatk_mark_duplicates",
        action="store_true",
        help="Use GATK MarkDuplicates.",
    )
    snv_ind_subparser.add_argument(
        "--use_dragen",
        action="store_true",
        help="Use DRAGEN.",
    )
    snv_ind_subparser.add_argument(
        "--output_folder",
        type=str,
        help="Output folder for results.",
    )
    snv_ind_subparser.add_argument(
        "--run_name",
        type=str,
        help="Run name.",
    )
    snv_ind_subparser.add_argument(
        "--ref_fasta",
        type=str,
        help="Path to the reference FASTA file.",
    )
    snv_ind_subparser.add_argument(
        "--ref_dict",
        type=str,
        help="Path to the reference dictionary file.",
    )
    snv_ind_subparser.add_argument(
        "--interval_list",
        type=str,
        help="Path to the interval list file.",
    )
    snv_ind_subparser.add_argument(
        "--scatter_count",
        type=int,
        help="Scatter count.",
    )
    snv_ind_subparser.add_argument(
        "--germline_resource",
        type=str,
        help="Path to the germline resource VCF.",
    )
    snv_ind_subparser.add_argument(
        "--index_image",
        type=str,
        help="Path to the index image.",
    )
    snv_ind_subparser.add_argument(
        "--downsampling_stride",
        type=int,
        help="Downsampling stride.",
    )
    snv_ind_subparser.add_argument(
        "--panel_of_normals",
        type=str,
        help="Panel of normals file.",
    )
    snv_ind_subparser.add_argument(
        "--variants_for_contamination",
        type=str,
        help="Variants for contamination file.",
    )
    snv_ind_subparser.add_argument(
        "--max_reads_per_alignment_start",
        type=int,
        help="Max reads per alignment start.",
    )
    snv_ind_subparser.add_argument(
        "--max_suspicious_reads_per_alignment_start",
        type=int,
        help="Max suspicious reads per alignment start.",
    )
    snv_ind_subparser.add_argument(
        "--max_population_af",
        type=float,
        help="Max population allele frequency.",
    )
    snv_ind_subparser.add_argument(
        "--lrom",
        action="store_true",
        help="Enable LROM.",
    )
    snv_ind_subparser.add_argument(
        "--interval_padding",
        type=int,
        help="Interval padding.",
    )
    snv_ind_subparser.add_argument(
        "--tumor_samples",
        type=str,
        help="JSON string or path to tumor samples list.",
    )
    snv_ind_subparser.add_argument(
        "--normal_samples",
        type=str,
        help="JSON string or path to normal samples list.",
    )
    # ------------------------------------ CNV SUBPARSER ------------------------------------
    cnv_config_subparser = subparsers.add_parser(
        "cnv",
        help="Run the CNV pipeline.",
        formatter_class=RichHelpFormatter,
    )
    cnv_config_subparser.add_argument(
        "--config",
        type=str,
        # required=True,
        help="Path to the configuration file for the CNV pipeline.",
    )
    cnv_config_subparser.add_argument(
        "--algotype",
        type=str,
        help="Algorithm type for BWA indexing.",
    )
    cnv_config_subparser.add_argument(
        "--index_fasta",
        action="store_true",
        help="Whether to index the FASTA file.",
    )
    cnv_config_subparser.add_argument(
        "--aligner_threads",
        type=int,
        help="Number of threads for the aligner.",
    )
    cnv_config_subparser.add_argument(
        "--remove_all_duplicates",
        action="store_true",
        help="Remove all duplicates.",
    )
    cnv_config_subparser.add_argument(
        "--remove_sequencing_duplicates",
        action="store_true",
        help="Remove sequencing duplicates.",
    )
    cnv_config_subparser.add_argument(
        "--use_gatk_mark_duplicates",
        action="store_true",
        help="Use GATK MarkDuplicates.",
    )
    cnv_config_subparser.add_argument(
        "--use_dragen",
        action="store_true",
        help="Use DRAGEN.",
    )
    cnv_config_subparser.add_argument(
        "--output_folder",
        type=str,
        help="Output folder for results.",
    )
    cnv_config_subparser.add_argument(
        "--run_name",
        type=str,
        help="Run name.",
    )
    cnv_config_subparser.add_argument(
        "--ref_fasta",
        type=str,
        help="Path to the reference FASTA file.",
    )
    cnv_config_subparser.add_argument(
        "--ref_dict",
        type=str,
        help="Path to the reference dictionary file.",
    )
    cnv_config_subparser.add_argument(
        "--interval_list",
        type=str,
        help="Path to the interval list file.",
    )
    cnv_config_subparser.add_argument(
        "--common_sites",
        type=str,
        help="Path to the common sites interval list.",
    )
    cnv_config_subparser.add_argument(
        "--pon",
        type=str,
        help="Panel of normals HDF5 file.",
    )
    cnv_config_subparser.add_argument(
        "--blacklist_intervals",
        type=str,
        help="Path to blacklist intervals file.",
    )
    cnv_config_subparser.add_argument(
        "--minimum_base_quality",
        type=int,
        help="Minimum base quality.",
    )
    cnv_config_subparser.add_argument(
        "--number_of_eigensamples",
        type=int,
        help="Number of eigensamples.",
    )
    cnv_config_subparser.add_argument(
        "--minimum_total_allele_count_case",
        type=int,
        help="Minimum total allele count (case).",
    )
    cnv_config_subparser.add_argument(
        "--minimum_total_allele_count_normal",
        type=int,
        help="Minimum total allele count (normal).",
    )
    cnv_config_subparser.add_argument(
        "--genotyping_homozygous_log_ratio_threshold",
        type=float,
        help="Genotyping homozygous log ratio threshold.",
    )
    cnv_config_subparser.add_argument(
        "--genotyping_base_error_rate",
        type=float,
        help="Genotyping base error rate.",
    )
    cnv_config_subparser.add_argument(
        "--maximum_number_of_segments_per_chromosome",
        type=int,
        help="Max segments per chromosome.",
    )
    cnv_config_subparser.add_argument(
        "--kernel_variance_copy_ratio",
        type=float,
        help="Kernel variance copy ratio.",
    )
    cnv_config_subparser.add_argument(
        "--kernel_variance_allele_fraction",
        type=float,
        help="Kernel variance allele fraction.",
    )
    cnv_config_subparser.add_argument(
        "--kernel_scaling_allele_fraction",
        type=float,
        help="Kernel scaling allele fraction.",
    )
    cnv_config_subparser.add_argument(
        "--kernel_approximation_dimension",
        type=int,
        help="Kernel approximation dimension.",
    )
    cnv_config_subparser.add_argument(
        "--window_size",
        nargs="+",
        type=int,
        help="Window sizes.",
    )
    cnv_config_subparser.add_argument(
        "--number_of_changepoints_penalty_factor",
        type=float,
        help="Changepoints penalty factor.",
    )
    cnv_config_subparser.add_argument(
        "--minor_allele_fraction_prior_alpha",
        type=float,
        help="Minor allele fraction prior alpha.",
    )
    cnv_config_subparser.add_argument(
        "--number_of_samples_copy_ratio",
        type=int,
        help="Number of samples (copy ratio).",
    )
    cnv_config_subparser.add_argument(
        "--number_of_burn_in_samples_copy_ratio",
        type=int,
        help="Burn-in samples (copy ratio).",
    )
    cnv_config_subparser.add_argument(
        "--number_of_samples_allele_fraction",
        type=int,
        help="Number of samples (allele fraction).",
    )
    cnv_config_subparser.add_argument(
        "--number_of_burn_in_samples_allele_fraction",
        type=int,
        help="Burn-in samples (allele fraction).",
    )
    cnv_config_subparser.add_argument(
        "--smoothing_credible_interval_threshold_copy_ratio",
        type=float,
        help="Smoothing credible interval threshold (copy ratio).",
    )
    cnv_config_subparser.add_argument(
        "--smoothing_credible_interval_threshold_allele_fraction",
        type=float,
        help="Smoothing credible interval threshold (allele fraction).",
    )
    cnv_config_subparser.add_argument(
        "--maximum_number_of_smoothing_iterations",
        type=int,
        help="Max smoothing iterations.",
    )
    cnv_config_subparser.add_argument(
        "--number_of_smoothing_iterations_per_fit",
        type=int,
        help="Smoothing iterations per fit.",
    )
    cnv_config_subparser.add_argument(
        "--neutral_segment_copy_ratio_lower_bound",
        type=float,
        help="Neutral segment copy ratio lower bound.",
    )
    cnv_config_subparser.add_argument(
        "--neutral_segment_copy_ratio_upper_bound",
        type=float,
        help="Neutral segment copy ratio upper bound.",
    )
    cnv_config_subparser.add_argument(
        "--outlier_neutral_segment_copy_ratio_z_score_threshold",
        type=float,
        help="Outlier neutral segment copy ratio z-score threshold.",
    )
    cnv_config_subparser.add_argument(
        "--calling_copy_ratio_z_score_threshold",
        type=float,
        help="Calling copy ratio z-score threshold.",
    )
    cnv_config_subparser.add_argument(
        "--minimum_contig_length",
        type=int,
        help="Minimum contig length.",
    )
    cnv_config_subparser.add_argument(
        "--maximum_copy_ratio",
        type=float,
        help="Maximum copy ratio.",
    )
    cnv_config_subparser.add_argument(
        "--point_size_copy_ratio",
        type=int,
        help="Point size (copy ratio).",
    )
    cnv_config_subparser.add_argument(
        "--point_size_allele_fraction",
        type=int,
        help="Point size (allele fraction).",
    )
    cnv_config_subparser.add_argument(
        "--tumor_samples",
        type=str,
        help="JSON string or path to tumor samples list.",
    )
    cnv_config_subparser.add_argument(
        "--normal_samples",
        type=str,
        help="JSON string or path to normal samples list.",
    )
    # ------------------------------------ GET CONFIG SUBPARSER ------------------------------------
    get_config_subparser = subparsers.add_parser(
        "get-config",
        help="Get examples pipeline configurations in json format.",
        formatter_class=RichHelpFormatter,
    )
    get_config_subparser.add_argument(
        "configs",
        nargs="*",
    )
    # PARSE ARGS
    argcomplete.autocomplete(parser, always_complete_options=True)
    args = parser.parse_args()

    # ============================ BWA DEDUP SUBPARSER ============================
    if args.command == "bwa_dedup":
        # BWA DEDUP required keys
        bwa_required_keys = [
            "ref_fasta",
            "bed_file",
            "out_path",
            "r1",
            "r2",
            "sample_name",
        ]
        # Load config if provided
        config = {}
        if args.config:
            with open(args.config) as f:
                config = json.load(f)

        # Overwrite defaults with config values only
        bwa_dedup_config.update({k: v for k, v in config.items() if v is not None})

        # Then overwrite with command line args
        for key, value in vars(args).items():
            if value is not None and key != "command":
                bwa_dedup_config[key] = value

        _dict_print(bwa_dedup_config, command=args.command)
        # Check for required keys
        not_found_keys = [key for key in bwa_required_keys if bwa_dedup_config[key] is None]
        if not_found_keys:
            _print_not_found_keys(not_found_keys)
        else:
            bwa_dedup(**bwa_dedup_config)  # acceptes kwargs as config

    # ============================ GATK GCNV SUBPARSER ============================
    elif args.command == "gatk_gcnv":
        # GATK GCNV required keys
        gatk_gcnv_required_keys = [
            "input_json",
            "output_dir",
            "exome_loc",
            "genome_fasta",
            "ref_dict",
            "contig_pp",
        ]
        # Load config if provided
        config = {}
        if args.config:
            with open(args.config) as f:
                config = json.load(f)

        # Overwrite defaults with config values only
        gatk_gcnv_config.update({k: v for k, v in config.items() if v is not None})

        # Then overwrite with command line args
        for key, value in vars(args).items():
            if value is not None and key != "command":
                gatk_gcnv_config[key] = value

        _dict_print(gatk_gcnv_config, command=args.command)
        # Check for required keys
        not_found_keys = [key for key in gatk_gcnv_required_keys if gatk_gcnv_config[key] is None]
        if not_found_keys:
            _print_not_found_keys(not_found_keys)
        else:
            gatk_gcnv(**gatk_gcnv_config)  # acceptes kwargs as config

    # ============================ PON SUBPARSER ============================
    elif args.command == "cnv_pon":
        # PON required keys
        pon_required_keys = [
            "output_folder",
            "ref_fasta",
            "ref_dict",
            "interval_list",
            "pon_id",
            "normal_bams",
            "blacklist_intervals",
        ]

        # Load config if provided
        config = {}
        if args.config:
            with open(args.config) as f:
                config = json.load(f)

        # Overwrite defaults with config values only
        cnv_pon_config.update({k: v for k, v in config.items() if v is not None})

        # Then overwrite with command line args
        for key, value in vars(args).items():
            if value is not None and key != "command":
                cnv_pon_config[key] = value

        _dict_print(cnv_pon_config, command=args.command)
        # Check for required keys
        not_found_keys = [key for key in pon_required_keys if cnv_pon_config[key] is None]
        if not_found_keys:
            _print_not_found_keys(not_found_keys)
        else:
            pon(**cnv_pon_config)  # acceptes kwargs as config

    # ============================ SNV/INDEL SUBPARSER ============================
    elif args.command == "snv_indel":
        # SNV/INDEL required keys
        snv_ind_required_keys = [
            "output_folder",
            "run_name",
            "ref_fasta",
            "ref_dict",
            "interval_list",
            "germline_resource",
            "index_image",
            "tumor_samples",
            "normal_samples",
        ]

        # Load config if provided
        config = {}
        if args.config:
            with open(args.config) as f:
                config = json.load(f)

        # Overwrite defaults with config values only
        snv_ind_config.update({k: v for k, v in config.items() if v is not None})

        # Then overwrite with command line args
        for key, value in vars(args).items():
            if value is not None and key != "command":
                snv_ind_config[key] = value

        _dict_print(snv_ind_config, command=args.command)
        # Check for required keys
        not_found_keys = [key for key in snv_ind_required_keys if snv_ind_config[key] is None]
        if not_found_keys:
            _print_not_found_keys(not_found_keys)
        else:
            snv_indel(snv_ind_config)  # accepts dictionary as config

    # ============================ CNV SUBPARSER ============================
    elif args.command == "cnv":
        # CNV required keys
        cnv_required_keys = [
            "output_folder",
            "run_name",
            "ref_fasta",
            "ref_dict",
            "interval_list",
            "common_sites",
            "pon",
            "blacklist_intervals",
            "tumor_samples",
            "normal_samples",
        ]

        # Load config if provided
        config = {}
        if args.config:
            with open(args.config) as f:
                config = json.load(f)

        # Overwrite defaults with config values only
        cnv_config.update({k: v for k, v in config.items() if v is not None})

        # Then overwrite with command line args
        for key, value in vars(args).items():
            if value is not None and key != "command":
                cnv_config[key] = value

        # Print the configuration dictionary
        _dict_print(cnv_config, command=args.command)

        # Check for required keys
        not_found_keys = [key for key in cnv_required_keys if cnv_config[key] is None]
        if not_found_keys:
            _print_not_found_keys(not_found_keys)
        else:
            somatic_CNV(cnv_config)  # accepts dictionary as config

    elif args.command == "get-config":
        if not args.configs:
            entry_dir = Path(__file__).parent.parent
            rprint("[bold green]Available configurations:[/bold green]")
            for file in (entry_dir / "configs").glob("*.json"):
                rprint(f"[grey85]{file.stem}[/grey85]")
        else:
            for config_name in args.configs:
                bring_config(config_name)

    else:
        print(parser.format_help())


if __name__ == "__main__":
    main()



import functools
import time


def timer(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        start = time.time()
        result = fn(*args, **kwargs)
        end = time.time()
        elapsed = end - start
        print(f"Finished in {elapsed:.2f} seconds...")

        return result

    return inner


@timer
def _test_timer():
    time.sleep(0.121341)


if __name__ == "__main__":
    _test_timer()



import os
import time
from concurrent.futures import ThreadPoolExecutor, as_completed

import pandas as pd

from bipackage.util.utilities import timer


# Keep the function outside the class
def run_bedtools_command(args):
    bam_file, exome_bait, input_dir = args
    sample_name = os.path.splitext(os.path.basename(bam_file))[0]
    output_name = sample_name + "_counts.tsv"
    command = f"bedtools multicov -bed {exome_bait} -bams {bam_file} -q 20 -p > {os.path.join(input_dir, output_name)}"
    os.system(command)


class CountsFromBam:
    def __init__(self, input_dir, exome_bait, num_threads):
        self.input_dir = input_dir
        self.exome_bait = exome_bait
        self.num_threads = num_threads

    def process_bams(self):
        bam_list = []

        # Find all BAM files recursively in the input directory
        for root, dirs, files in os.walk(self.input_dir):
            for file in files:
                if file.endswith(".bam"):
                    bam_list.append(os.path.join(root, file))

        # Prepare arguments for the bedtools command
        arguments = [(bam_file, self.exome_bait, self.input_dir) for bam_file in bam_list]

        # Use ThreadPoolExecutor to parallelize the bedtools command execution
        with ThreadPoolExecutor(max_workers=self.num_threads) as executor:
            futures = [executor.submit(run_bedtools_command, arg) for arg in arguments]

            for future in as_completed(futures):
                try:
                    future.result()  # To catch exceptions raised during thread execution
                except Exception as e:
                    print(f"Error occurred: {e}")

    def create_count_matrix(self):
        counts_paths = [
            os.path.join(self.input_dir, item) for item in os.listdir(self.input_dir) if item.endswith("_counts.tsv")
        ]
        file_names = [item for item in os.listdir(self.input_dir) if item.endswith("_counts.tsv")]

        # Check if count files exist
        if not counts_paths:
            print("No count files found!")
            return

        counts_names = [name.split("_counts")[0] for name in file_names]

        # Read the first file and initialize the DataFrame
        count_mat = pd.read_csv(counts_paths[0], sep="\t", header=None)

        count_mat.columns = ["chromosome", "start", "end", counts_names[0]]

        # Append data from the remaining files
        for i in range(1, len(counts_paths)):
            df = pd.read_csv(counts_paths[i], sep="\t", header=None)
            con_df = df.iloc[:, 3]
            con_df.name = counts_names[i]
            count_mat = pd.concat([count_mat, con_df], axis=1)

        # Save the final count matrix
        count_mat.to_csv(os.path.join(self.input_dir, "Counts_matrix.txt"), sep="\t", index=False)


@timer
def bam_counts(input_dir: str, exome_bait: str, num_threads: int = 40) -> None:
    """
    Get counts from BAM file.

    Parameters
    ----------
    input_dir : str
        Input directory.
    exome_bait : str
        Exom bait.
    num_threads : int
        Number of threads to use, default is 40.

    Returns
    -------
    None
    """
    counts = CountsFromBam(input_dir, exome_bait, num_threads)
    counts.process_bams()
    # Create count matrix
    counts.create_count_matrix()
    return


if __name__ == "__main__":
    pass



import argparse
import os

import pandas as pd

from bipackage.util.utilities import timer


'''Script can compile detailed bam statistics produced by Megabolt'''

def parse_coverage_report(file_path):
    """Parse a .coverage.report file and return a dictionary of its contents."""
    data = {}

    sample_name = os.path.basename(file_path).split(".")[0]  # Extract sample name from filename
    data["Sample"] = sample_name

    with open(file_path, "r") as f:
        lines = f.readlines()
        for line in lines[2:]:  # Skip the first two lines
            parts = line.strip().split()  # Splitting based on whitespace
            if len(parts) >= 2:
                key = " ".join(parts[:-1])  # Key is everything except the last element
                value = parts[-1]  # Value is the last element
                data[key] = str(value)  # Store values as strings to prevent conflicts

    return data


def find_coverage_report_files(root_dir):
    """Recursively find all .coverage.report files in a directory."""
    report_files = []
    for root, _, files in os.walk(root_dir):
        for file in files:
            if file.endswith(".coverage.report"):
                report_files.append(os.path.join(root, file))
    return report_files


def compile_reports_to_csv(root_dir: str, output_csv: str):
    """Compile all .coverage.report files into a single CSV."""
    report_files = find_coverage_report_files(root_dir)
    data_list = [parse_coverage_report(file) for file in report_files]
    df = pd.DataFrame(data_list)
    df.to_csv(output_csv, index=False)
    print(f"Compiled {len(report_files)} files into {output_csv}")


@timer
def compile_bam_stats(root_directory: str, output_csv: str) -> None:
    """
    Complies BAM stats to CSV.

    Parameters
    ----------
    root_directory : str
        Root directory for the operations.
    output_csv : str
        Output csv file
    """
    compile_reports_to_csv(root_dir=root_directory, output_csv=output_csv)


if __name__ == "__main__":
    pass



import logging
import os
import subprocess

import pandas as pd

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")


class Indexer:
    def __init__(self, reference_fasta: str, algotype: str):
        self.reference_fasta = reference_fasta
        self.algotype = algotype

    def index(self):
        logging.info("bwa - Indexing the reference genome")
        command = f"bwa index -a {self.algotype} {self.reference_fasta}"

        try:
            os.system(command)
            logging.info("bwa - Successfully indexed the reference genome")
        except Exception as e:
            logging.error(f"bwa - An error occurred while indexing the reference genome: {e}")


class Mapper:
    def __init__(
        self,
        reference_fasta: str,
        sample_id: str,
        r1: str,
        r2: str,
        threads: int,
        output_path: str,
        use_dragen: bool = False,
        hash_table: str = None,
    ):
        self.reference_fasta = reference_fasta
        self.sample_id = sample_id
        self.r1 = r1
        self.r2 = r2
        self.threads = threads
        self.output_path = output_path
        self.use_dragen = use_dragen
        self.hash_table = hash_table

    def map(self):
        sample_bam_path = os.path.join(self.output_path, self.sample_id)
        os.makedirs(sample_bam_path, exist_ok=True)
        sam_file = self.sample_id + ".sam"

        if self.use_dragen:
            logging.info(f"dragen - Mapping {self.r1} and {self.r2} to the reference genome")
            command = (
                f"dragen-os -r {self.hash_table} "
                f"-1 {self.r1} -2 {self.r2} "
                f"--RGSM {self.sample_id} --RGID {self.sample_id} "
                f"--num-threads {self.threads} > {os.path.join(sample_bam_path, sam_file)}"
            )

            try:
                os.system(command)
                logging.info(f"dragen - Successfully mapped {self.r1} and {self.r2} to the reference genome")
            except Exception as e:
                logging.error(f"dragen - An error occurred while mapping {self.r1} and {self.r2}: {e}")

        else:
            logging.info(f"bwa - Mapping {self.r1} and {self.r2} to the reference genome")
            command = (
                f"bwa mem -M -t {self.threads} "
                f"-R '@RG\\tID:{self.sample_id}\\tSM:{self.sample_id}\\tLB:Mylib\\tPU:Illumina' "
                f"{self.reference_fasta} {self.r1} {self.r2} > {os.path.join(sample_bam_path, sam_file)}"
            )

            try:
                os.system(command)
                logging.info(f"bwa - Successfully mapped {self.r1} and {self.r2} to the reference genome")
            except Exception as e:
                logging.error(f"bwa - An error occurred while mapping {self.r1} and {self.r2}: {e}")

            return os.path.join(sample_bam_path, sam_file)


class BAMConverter:
    def __init__(
        self,
        sample_id,
        sam_path,
        threads,
        use_mark_duplicates,
        remove_all_duplicates,
        remove_sequencing_duplicates,
    ):
        self.sample_id = sample_id
        self.sam_path = sam_path
        self.threads = threads
        self.use_mark_duplicates = use_mark_duplicates
        self.remove_all_duplicates = remove_all_duplicates
        self.remove_sequencing_duplicates = remove_sequencing_duplicates
        self.out_path = os.path.dirname(self.sam_path)
        self.filtered_bam_file = self.sample_id + ".bam"
        self.filtered_bam = os.path.join(self.out_path, self.filtered_bam_file)
        self.metrics_file = self.sample_id + "_dup_metrics.txt"
        self.metrics_path = os.path.join(self.out_path, self.metrics_file)
        self.sorted_bam_file = "sorted_" + self.sample_id + ".bam"
        self.sorted_bam = os.path.join(self.out_path, self.sorted_bam_file)

    def sort_bam(self):
        sort_cmd = f"samtools sort -@ {self.threads} {self.sam_path} -o {self.sorted_bam}"

        try:
            os.system(sort_cmd)
            logging.info(f"Samtools - Successfully sorted bam file for {self.sample_id} > {self.sorted_bam}")
        except Exception as e:
            logging.error(f"Samtools - An error occurred while sorting bam file for {self.sample_id}: {e}")

    def mark_duplicates(self):
        if self.use_mark_duplicates:
            logging.info(f"MarkDuplicatesSpark - Marking and removing duplicates for bam file of {self.sample_id}")
            marker_cmd = f"gatk MarkDuplicatesSpark -I {self.sorted_bam} -O {self.filtered_bam} --metrics-file {self.metrics_path} --spark-master local[{self.threads}]"

            if self.remove_sequencing_duplicates:
                marker_cmd += " --remove-sequencing-duplicates"

            if self.remove_all_duplicates:
                marker_cmd += " --remove-all-duplicates"

            try:
                os.system(marker_cmd)
                logging.info(
                    f"MarkDuplicatesSpark - Successfully marked and removed duplicates for {self.sample_id} > {self.filtered_bam}"
                )
            except Exception as e:
                logging.error(
                    f"An error occurred while marking and removing duplicates for bam file of {self.sample_id}: {e}"
                )
        else:
            logging.info(f"Sambamba - Marking and removing duplicates for bam file of {self.sample_id}")
            marker_cmd = f"sambamba markdup -t {self.threads} {self.sorted_bam} {self.filtered_bam}"

            if self.remove_all_duplicates:
                marker_cmd += " --remove-duplicates"

            try:
                os.system(marker_cmd)
                logging.info(
                    f"Sambamba - Successfully marked and removed duplicates for {self.sample_id} > {self.filtered_bam}"
                )
            except Exception as e:
                logging.error(
                    f"An error occurred while marking and removing duplicates for bam file of {self.sample_id}: {e}"
                )

        return self.filtered_bam

    def index_bam(self):
        index_cmd = f"samtools index -@ {self.threads} {self.filtered_bam}"

        try:
            os.system(index_cmd)
            logging.info(f"Samtools - Successfully indexed bam file for {self.sample_id} > {self.filtered_bam}")
        except Exception as e:
            logging.error(f"Samtools - An error occurred while indexing bam file for {self.sample_id}: {e}")

    def clean_up(self):
        try:
            os.remove(self.sam_path)
            os.remove(self.sorted_bam)
            logging.info(f"Clean up - Successfully removed intermediate files for {self.sample_id}")
        except Exception as e:
            logging.error(f"Clean up - An error occurred while removing intermediate files for {self.sample_id}: {e}")

    def convert_markdedup(self):
        self.sort_bam()
        resulting_bam = self.mark_duplicates()
        self.index_bam()
        self.clean_up()

        return resulting_bam, self.out_path


class CoverageAnalysis:
    def __init__(self, reference, bam, sample_id, output_dir, bed, max_worker=40):
        self.reference = reference
        self.bed = bed
        self.bam = bam
        self.output_dir = output_dir
        self.max_worker = max_worker
        self.sample_name = sample_id

    def run_mosdepth(self):
        output = os.path.join(self.output_dir, self.sample_name)

        command = [
            "mosdepth",
            output,
            f"-x -Q 1 -t {self.max_worker}",
            "--by",
            self.bed,
            f"-f {self.reference}",
            f"-n -T 1,5,10,15,20,30,50,100,500,1500",
            self.bam,
        ]
        subprocess.run(" ".join(command), shell=True)

        thresholds = os.path.join(self.output_dir, self.sample_name + ".thresholds.bed.gz")
        summary = os.path.join(self.output_dir, self.sample_name + ".mosdepth.summary.txt")

        return thresholds, summary

    def calculate_coverages(self, qc_thresholds_file):
        result = []
        df = pd.read_csv(qc_thresholds_file, compression="gzip", header=None, sep="\t")
        columns = df.iloc[0].to_list()
        df = df.drop(df.index[0])
        df.columns = columns
        df["size"] = df["end"].astype(int) - df["start"].astype(int)

        total_size = df["size"].sum()
        print(total_size)

        # Coverage columns
        coverage_columns = [col for col in df.columns if col.endswith("X")]

        # Calculate the sum of each coverage column and then calculate coverage
        coverage_dict = {}
        for col in coverage_columns:
            df[col] = pd.to_numeric(df[col], errors="coerce").astype(int)
            coverage_sum = df[col].sum()
            coverage = coverage_sum / total_size
            coverage_dict[col] = coverage

        # Display the coverage values
        for col, cov in coverage_dict.items():
            result.append(
                {
                    "key": f"PCT of QC coverage region with coverage [{col.lower()}: inf):",
                    "value": round(cov * 100, 2),
                }
            )

        coverage_keys = list(coverage_dict.keys())
        for i in range(len(coverage_keys) - 1):
            diff = df[coverage_keys[i]] - df[coverage_keys[i + 1]]
            result.append(
                {
                    "key": f"PCT of QC coverage region with coverage [{coverage_keys[i].lower()}:{coverage_keys[i + 1].lower()}):",
                    "value": round((diff.sum() / total_size) * 100, 2),
                }
            )

        return result

    def OnTargetContigs(self, depth_summary):
        df = pd.read_csv(depth_summary, sep="\t")
        df = df.iloc[:50]
        # Select values without _region. all chromosomes
        df1 = df[~df["chrom"].str.endswith("_region")]
        total_bases = df1["bases"].astype(int).sum()

        # Select target regions only
        df = df[df["chrom"].str.endswith("_region")]
        df["chrom"] = df["chrom"].str.replace("_region", "")
        target_total_bases = df["bases"].astype(int).sum()
        target_length = df["length"].astype(int).sum()
        mean_coverage = round(target_total_bases / target_length, 2)

        df = df.drop(["length", "bases", "min", "max"], axis=1)

        df.columns = ["key", "value"]
        # uniformity_2 = sum(1 for x in total if x > 0.2 * mean_coverage) / total_bases * 100
        # uniformity_4 = sum(1 for x in total if x > 0.4 * mean_coverage) / total_bases * 100
        final_result = [
            {"key": "Aligned bases", "value": total_bases},
            {"key": "Aligned bases in QC coverage region", "value": target_total_bases},
            {
                "key": "Average alignment coverage over QC coverage region",
                "value": mean_coverage,
            },
            # {"key": "Uniformity of coverage (PCT > 0.2*mean)", "value": uniformity_2},
            # {"key": "Uniformity of coverage (PCT > 0.4*mean)", "value": uniformity_4},
            {"key": "PCT of QC coverage region with coverage [0x: inf):", "value": 100},
        ]
        return df.to_dict(orient="records"), final_result

    def process(self):
        try:
            logging.info(f"Starting process for sample: {self.sample_name}")

            qc_thresholds_file, depth_summary_file = self.run_mosdepth()
            logging.info(f"Generated Mosdepth output: thresholds: {qc_thresholds_file}, summary: {depth_summary_file}")

            if os.path.exists(qc_thresholds_file):
                logging.info(f"Found thresholds file: {qc_thresholds_file}, proceeding to calculate coverages")
                metrics = self.calculate_coverages(qc_thresholds_file)
            else:
                logging.error(f"Thresholds file not found: {qc_thresholds_file}")
                return

            if os.path.exists(depth_summary_file):
                logging.info(f"Found summary file: {depth_summary_file}, proceeding to OnTargetContigs")
                _, final_results = self.OnTargetContigs(depth_summary_file)
                metrics.extend(final_results)
            else:
                logging.error(f"Summary file not found: {depth_summary_file}")
                return

            output_metrics_file = os.path.join(self.output_dir, f"{self.sample_name}_metrics.txt")
            logging.info(f"Writing metrics to file: {output_metrics_file}")

            with open(output_metrics_file, "w") as f:
                for metric in metrics:
                    f.write(f"{metric['key']} {metric['value']}\n")
            logging.info(f"Successfully wrote metrics for {self.sample_name}")

        except Exception as e:
            logging.error(
                f"An error occurred while processing {self.sample_name}: {e}",
                exc_info=True,
            )


# TODO: MOVE
def bwa_dedup(
    ref_fasta:str,
    bed_file:str,
    out_path:str,
    r1:str,
    r2:str,
    sample_name:str,
    *,
    at:str="bwtsw",
    threads:int=40,
    remove_dups:bool=False,
    remove_seq_dups:bool=False,
    use_md=False,
) -> None:
    """
    Run BWA mapping, deduplication, and coverage analysis using Mosdepth.

    Parameters
    ----------
    ref_fasta : str
        Path to the reference FASTA file.
    bed_file : str
        Path to the BED file for coverage analysis.
    out_path : str
        Output directory for results.
    r1 : str
        Path to the first read FASTQ file.
    r2 : str
        Path to the second read FASTQ file.
    sample_name : str
        Sample identifier for output files.
    at : str, optional
        Algorithm type for BWA indexing, by default 'bwtsw'.
    threads : int, optional
        Number of threads to use for processing, by default 40.
    remove_dups : bool, optional
        Whether to remove all duplicates, by default False.
    remove_seq_dups : bool, optional
        Whether to remove sequencing duplicates, by default False.
    use_md : bool, optional
        Whether to use MarkDuplicatesSpark, by default False.
    """
    # indexit = Indexer(reference_fasta=ref_fasta, algotype=at) # use once
    # indexit.index() # use once

    mapit = Mapper(
        reference_fasta=ref_fasta,
        sample_id=sample_name,
        r1=r1,
        r2=r2,
        output_path=out_path,
        threads=threads,
    )
    sam_file = mapit.map()

    dedup = BAMConverter(
        sam_path=sam_file,
        sample_id=sample_name,
        threads=threads,
        remove_all_duplicates=remove_dups,
        remove_sequencing_duplicates=remove_seq_dups,
        use_mark_duplicates=use_md,
    )
    final_bam, out_path_metrics = dedup.convert_markdedup()

    coverage_analysis = CoverageAnalysis(
        reference=ref_fasta,
        bam=final_bam,
        sample_id=sample_name,
        output_dir=out_path_metrics,
        bed=bed_file,
        max_worker=threads,
    )
    coverage_analysis.process()

    return


import json
import os
import time


# subcommand 1
def is_mounted(folder_path: str):
    result: bool = os.path.ismount(folder_path)
    print(f"{folder_path} is mounted: {result}")
    return result


# subcommand 2
def mount_server(username: str, server_address: str, mount_folder: str, password: str, version=None):
    """Mount nas server."""
    if version:
        mount_command = f"sudo mount -t cifs //{server_address} {mount_folder} -o username={username},password={password},vers={version}"
    else:
        mount_command = (
            f"sudo mount -t cifs //{server_address} {mount_folder} -o username={username},password={password}"
        )

    os.system(mount_command)
    return


# subcommand 3
def check_reconnect(config_file: str):
    """Check reconnects."""
    with open(config_file, "r") as file:
        config_file = json.load(file)

    for key, value in config_file.items():
        if not is_mounted(key):
            print(f"Folder {key} is not mounted. Reconnecting...")
            username = config_file.get(key).get("username")
            server_address = config_file.get(key).get("server_address")
            version = config_file.get(key).get("version")
            password = config_file.get(key).get("password")

            mount_server(username, server_address, key, password, version)

            time.sleep(5)

            if is_mounted(key):
                print(f"Successfully connected {server_address} to {key}")
            else:
                print(f"Failed to connect to {server_address}")
        else:
            print(f"Folder {key} is already mounted")

    return



import gzip
import logging
import os
import subprocess
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor

from bipackage.util.utilities import timer


class FastqMerger:
    def __init__(self, input_folder_paths, output_folder, names):
        self.input_folder_paths = input_folder_paths
        self.output_folder = output_folder
        self.names = names

    def process_samples(self, name):
        r1_paths, r2_paths = [], []

        for folder in self.input_folder_paths:
            for root, _, files in os.walk(folder):
                for filename in files:
                    if (
                        name.lower().strip() in filename.lower().strip()
                        and (filename.endswith("R1_001.fastq.gz") or filename.endswith("2.fq.gz"))
                        and "_MF_R2" not in filename
                    ):
                        r1_paths.append(os.path.join(root, filename))
                    if (
                        name.lower().strip() in filename.lower().strip()
                        and (filename.endswith("R2_001.fastq.gz") or filename.endswith("2.fq.gz"))
                        and "_MF_R2" not in filename
                    ):
                        r2_paths.append(os.path.join(root, filename))
        return r1_paths, r2_paths

    def merge_fastq_files_command(self, input_paths, output_path):
        return f"zcat {' '.join(input_paths)} | gzip > {output_path}"

    def prepare_merge_commands(self):
        commands = []
        for name in self.names:
            print(f"#################{name}###################")

            r1_paths, r2_paths = self.process_samples(name)

            fq1_output_path = os.path.join(self.output_folder, f"{name}_MF_R1_001.fastq.gz")
            fq2_output_path = os.path.join(self.output_folder, f"{name}_MF_R2_001.fastq.gz")

            if r1_paths:
                commands.append(self.merge_fastq_files_command(r1_paths, fq1_output_path))
                print("###FOR R1###")
                print(self.merge_fastq_files_command(r1_paths, fq1_output_path))
            if r2_paths:
                commands.append(self.merge_fastq_files_command(r2_paths, fq2_output_path))
                print("###FOR R2###")
                print(self.merge_fastq_files_command(r2_paths, fq2_output_path))
        return commands

    @staticmethod
    def run_merge_command(command):
        result = subprocess.run(command, shell=True, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        stderr_output = result.stderr.decode("utf-8").strip()
        stdout_output = result.stdout.decode("utf-8").strip()

        if result.returncode != 0:
            logging.error(f"Merge - An error occurred while running command: {stderr_output}")
        else:
            logging.info(f"Merge - Command completed successfully: {stdout_output}")

        return

    def execute_merge(self, commands):
        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(self.run_merge_command, command) for command in commands]
            for future in futures:
                try:
                    future.result()
                except Exception as e:
                    logging.error(f"Error occurred during merging: {e}")

        return


@timer
def merge_it(*, folder_paths: list[str], sample_names: list[str], output_path: str) -> None:
    """
    Merge fastq files.

    Parameters
    ----------
    folder_paths : list[str]
        List of paths to folders.
    sample_names: list[str]
        List of name of the files.
    output_path : str
        Path to the output directory.

    Returns
    -------
    None
    """
    merger = FastqMerger(input_folder_paths=folder_paths, names=sample_names, output_folder=output_path)
    merge_commands = merger.prepare_merge_commands()
    merger.execute_merge(merge_commands)
    return


if __name__ == "__main__":
    pass



from bipackage.src import (
    bam_counts,
    bedfilegenerator,
    check_gzip_validity,  # DONE
    check_reconnect,  # DONE
    compile_bam_stats,  # DONE
    downsample,
    fastq_read_counter,  # DONE
    fastqvalidate,  # DONE
    is_mounted,  # DONE
    md5sumchecker,  # DONE
    merge_it,
    mount_server,  # DONE
    nipt_bcl2fastq,
    panelgenequery,  # DONE
    remove_undetermined_fastq,  # DONE
    undetermined_demultiplexer,
)

__all__ = [
    "bam_counts",
    "bedfilegenerator",
    "check_gzip_validity",
    "check_reconnect",
    "compile_bam_stats",
    "downsample",
    "fastq_read_counter",
    "fastqvalidate",
    "is_mounted",
    "md5sumchecker",
    "merge_it",
    "mount_server",
    "nipt_bcl2fastq",
    "panelgenequery",
    "remove_undetermined_fastq",
    "undetermined_demultiplexer",
]



import argparse
import csv
import json
import logging
import os
import subprocess

'''Script downsamples fastqs'''

import pandas as pd

from bipackage.util.utilities import timer

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")


class Indexer:
    def __init__(self, reference_fasta: str, algotype: str):
        self.reference_fasta = reference_fasta
        self.algotype = algotype

    def index(self):
        logging.info("bwa - Indexing the reference genome")
        command = f"bwa index -a {self.algotype} {self.reference_fasta}"

        try:
            os.system(command)
            logging.info("bwa - Successfully indexed the reference genome")
        except Exception as e:
            logging.error(f"bwa - An error occurred while indexing the reference genome: {e}")


class Mapper:
    def __init__(
        self,
        reference_fasta: str,
        sample_id: str,
        r1: str,
        r2: str,
        threads: int,
        output_path: str,
        use_dragen: bool = False,
        hash_table: str = None,
    ):
        self.reference_fasta = reference_fasta
        self.sample_id = sample_id
        self.r1 = r1
        self.r2 = r2
        self.threads = threads
        self.output_path = output_path
        self.use_dragen = use_dragen
        self.hash_table = hash_table

    def map(self):
        sample_bam_path = os.path.join(self.output_path, self.sample_id)
        os.makedirs(sample_bam_path, exist_ok=True)
        sam_file = self.sample_id + ".sam"

        if self.use_dragen:
            logging.info(f"dragen - Mapping {self.r1} and {self.r2} to the reference genome")
            command = (
                f"dragen-os -r {self.hash_table} "
                f"-1 {self.r1} -2 {self.r2} "
                f"--RGSM {self.sample_id} --RGID {self.sample_id} "
                f"--num-threads {self.threads} > {os.path.join(sample_bam_path, sam_file)}"
            )

            try:
                os.system(command)
                logging.info(f"dragen - Successfully mapped {self.r1} and {self.r2} to the reference genome")
            except Exception as e:
                logging.error(f"dragen - An error occurred while mapping {self.r1} and {self.r2}: {e}")

        else:
            logging.info(f"bwa - Mapping {self.r1} and {self.r2} to the reference genome")
            command = (
                f"bwa mem -M -t {self.threads} "
                f"-R '@RG\\tID:{self.sample_id}\\tSM:{self.sample_id}\\tLB:Mylib\\tPU:Illumina' "
                f"{self.reference_fasta} {self.r1} {self.r2} > {os.path.join(sample_bam_path, sam_file)}"
            )

            try:
                os.system(command)
                logging.info(f"bwa - Successfully mapped {self.r1} and {self.r2} to the reference genome")
            except Exception as e:
                logging.error(f"bwa - An error occurred while mapping {self.r1} and {self.r2}: {e}")

            return os.path.join(sample_bam_path, sam_file)


class BAMConverter:
    def __init__(
        self,
        sample_id,
        sam_path,
        threads,
        use_mark_duplicates,
        remove_all_duplicates,
        remove_sequencing_duplicates,
    ):
        self.sample_id = sample_id
        self.sam_path = sam_path
        self.threads = threads
        self.use_mark_duplicates = use_mark_duplicates
        self.remove_all_duplicates = remove_all_duplicates
        self.remove_sequencing_duplicates = remove_sequencing_duplicates
        self.out_path = os.path.dirname(self.sam_path)
        self.bam_file = self.sample_id + ".bam"
        self.output_bam = os.path.join(self.out_path, self.bam_file)
        self.filtered_bam_file = "filtered_" + self.sample_id + ".bam"
        self.filtered_bam = os.path.join(self.out_path, self.filtered_bam_file)
        self.metrics_file = self.sample_id + "_dup_metrics.txt"
        self.metrics_path = os.path.join(self.out_path, self.metrics_file)
        self.sorted_bam_file = "sorted_" + self.filtered_bam_file
        self.sorted_bam = os.path.join(self.out_path, self.sorted_bam_file)

    def create_bam_file(self):
        logging.info(f"Samtools - Creating a bam file for {self.sample_id} from sam file")
        converter_cmd = f"samtools view -@ {self.threads} -bS -o {self.output_bam} {self.sam_path}"

        try:
            os.system(converter_cmd)
            logging.info(f"Samtools - Successfully created bam file for {self.sample_id}")
        except Exception as e:
            logging.error(f"Samtools - An error occurred while creating bam file for {self.sample_id}: {e}")

    def mark_duplicates(self):
        if self.use_mark_duplicates:
            logging.info(f"MarkDuplicatesSpark - Marking and removing duplicates for bam file of {self.sample_id}")
            marker_cmd = f"gatk MarkDuplicatesSpark -I {self.output_bam} -O {self.filtered_bam} --metrics-file {self.metrics_path} --spark-master local[{self.threads}]"

            if self.remove_sequencing_duplicates:
                marker_cmd += " --remove-sequencing-duplicates"

            if self.remove_all_duplicates:
                marker_cmd += " --remove-all-duplicates"

            try:
                os.system(marker_cmd)
                logging.info(
                    f"MarkDuplicatesSpark - Successfully marked and removed duplicates for {self.sample_id} > {self.filtered_bam}"
                )
            except Exception as e:
                logging.error(
                    f"An error occurred while marking and removing duplicates for bam file of {self.sample_id}: {e}"
                )
        else:
            logging.info(f"Sambamba - Marking and removing duplicates for bam file of {self.sample_id}")
            marker_cmd = f"sambamba markdup -t {self.threads} {self.output_bam} {self.filtered_bam}"

            if self.remove_all_duplicates:
                marker_cmd += " --remove-duplicates"

            try:
                os.system(marker_cmd)
                logging.info(
                    f"Sambamba - Successfully marked and removed duplicates for {self.sample_id} > {self.filtered_bam}"
                )
            except Exception as e:
                logging.error(
                    f"An error occurred while marking and removing duplicates for bam file of {self.sample_id}: {e}"
                )

    def sort_bam(self):
        sort_cmd = f"samtools sort -@ {self.threads} {self.filtered_bam} -o {self.sorted_bam}"

        try:
            os.system(sort_cmd)
            logging.info(f"Samtools - Successfully sorted bam file for {self.sample_id} > {self.sorted_bam}")
        except Exception as e:
            logging.error(f"Samtools - An error occurred while sorting bam file for {self.sample_id}: {e}")

    def index_bam(self):
        index_cmd = f"samtools index -@ {self.threads} {self.sorted_bam}"

        try:
            os.system(index_cmd)
            logging.info(f"Samtools - Successfully indexed bam file for {self.sample_id} > {self.sorted_bam}")
        except Exception as e:
            logging.error(f"Samtools - An error occurred while indexing bam file for {self.sample_id}: {e}")

    def clean_up(self):
        try:
            os.remove(self.sam_path)
            os.remove(self.output_bam)
            os.remove(self.filtered_bam)
            logging.info(f"Clean up - Successfully removed intermediate files for {self.sample_id}")
        except Exception as e:
            logging.error(f"Clean up - An error occurred while removing intermediate files for {self.sample_id}: {e}")

    def convert_markdedup(self):
        self.create_bam_file()
        self.mark_duplicates()
        self.sort_bam()
        self.index_bam()
        self.clean_up()

        return self.sorted_bam


# class CoverageAnalysis:
#     def __init__(self, reference, bam, bed, output_dir, max_worker):
#         self.reference = reference
#         self.bam = bam
#         self.bed = bed
#         self.output_dir = output_dir
#         self.max_worker = max_worker

#     def run_mosdepth(self):

#         output_prefix = os.path.splitext(self.bam)[0]
#         output = os.path.join(self.output_dir, output_prefix)
#         command = [
#             "mosdepth", output, f"-x -Q 1 -t {self.max_worker}",
#             "--by", self.bed, f"-f {self.reference}",
#             f"-n -T 1,5,10,15,20,30,50,100,500,1500",
#             self.bam
#         ]
#         subprocess.run(" ".join(command), shell=True)

#     def calculate_coverages(self):

#         output_prefix = os.path.splitext(self.bam)[0]
#         qc_thresholds_file = os.path.join(self.output_dir, f"{output_prefix}.thresholds.bed.gz")

#         result = []

#         df = pd.read_csv(qc_thresholds_file, compression='gzip', header=None, sep='\t')
#         columns = df.iloc[0].to_list()
#         df = df.drop(df.index[0])
#         df.columns = columns
#         df['size'] = df['end'].astype(int) - df['start'].astype(int)

#         total_size = df['size'].sum()
#         print(total_size)

#         # Coverage columns
#         coverage_columns = [col for col in df.columns if col.endswith('X')]

#         # Calculate the sum of each coverage column and then calculate coverage
#         coverage_dict = {}
#         for col in coverage_columns:
#             df[col] = pd.to_numeric(df[col], errors='coerce').astype(int)
#             coverage_sum = df[col].sum()
#             coverage = coverage_sum / total_size
#             coverage_dict[col] = coverage

#         # Display the coverage values
#         for col, cov in coverage_dict.items():
#             result.append({"key": f"PCT of QC coverage region with coverage [{col.lower()}: inf):", "value": round(cov * 100, 2)})

#         coverage_keys = list(coverage_dict.keys())
#         for i in range(len(coverage_keys) - 1):
#             diff = df[coverage_keys[i]] - df[coverage_keys[i + 1]]
#             result.append({"key": f"PCT of QC coverage region with coverage [{coverage_keys[i].lower()}:{coverage_keys[i + 1].lower()}):", "value": round((diff.sum() / total_size) * 100, 2)})

#         return result

#     def OnTargetContigs(self):

#         output_prefix = os.path.splitext(self.bam)[0]
#         depth_summary = os.path.join(self.output_dir, f"{output_prefix}.mosdepth.summary.txt")

#         df = pd.read_csv(depth_summary, sep="\t")
#         df = df.iloc[:50]
#         # Select values without _region. all chromosomes
#         df1 = df[~df['chrom'].str.endswith('_region')]
#         total_bases = df1['bases'].astype(int).sum()

#         # Select target regions only
#         df = df[df['chrom'].str.endswith('_region')]
#         df['chrom'] = df['chrom'].str.replace('_region', '')
#         target_total_bases = df['bases'].astype(int).sum()
#         target_length = df['length'].astype(int).sum()
#         mean_coverage = round(target_total_bases / target_length, 2)

#         df = df.drop(['length', 'bases', 'min', 'max'], axis=1)

#         df.columns = ['key', 'value']
#         # uniformity_2 = sum(1 for x in total if x > 0.2 * mean_coverage) / total_bases * 100
#         # uniformity_4 = sum(1 for x in total if x > 0.4 * mean_coverage) / total_bases * 100
#         final_result = [
#             {"key": "Aligned bases", "value": total_bases},
#             {"key":"Aligned bases in QC coverage region", "value":target_total_bases},
#             {"key": "Average alignment coverage over QC coverage region", "value": mean_coverage},
#             # {"key": "Uniformity of coverage (PCT > 0.2*mean)", "value": uniformity_2},
#             # {"key": "Uniformity of coverage (PCT > 0.4*mean)", "value": uniformity_4},
#             {"key": "PCT of QC coverage region with coverage [0x: inf):", "value": 100}
#         ]

#         print(df.to_dict(orient="records"), final_result)

#         return df.to_dict(orient="records"), final_result


class Downsampler:
    def __init__(self, alignment_file, keep, sample_id, strategy, refseq):
        self.alignment_file = alignment_file
        self.sample_id = sample_id
        self.keep = keep
        self.strategy = strategy
        self.refseq = refseq
        # self.exome_bait = exome_bait
        # self.quality_threshold = quality_threshold
        # self.read_length = read_length
        # self.desired_mean_coverage = desired_mean_coverage
        self.bam_dir = os.path.dirname(self.alignment_file)

    # def get_bam_counts(self):

    #     sample_name = os.path.splitext(bam_file)[0]
    #     output_name = sample_name + "_counts.tsv"
    #     command = f"bedtools multicov -bed {self.exome_bait} -bams {self.alignment_file} -q {self.quality_threshold} -p > {os.path.join(self.bam_dir, output_name)}"

    #     try:
    #         os.system(command)
    #         logging.info("multicov - Successfully get the read counts the bam")
    #     except Exception as e:
    #         logging.error(f"multicov - An error occurred getting read counts from the bam: {e}")

    #     return os.path.join(self.bam_dir, output_name)

    # def calculate_keep(self, counts_file):

    #     counts_df = pd.read_csv(counts_file, sep="\t", header=None)

    #     counts_df = counts_df.iloc[:, [0,1,2,-1]]

    #     counts_df.columns = ["chrom", "start", "end", "counts"]

    #     counts_df["bait_length"] = (counts_df["end"] - counts_df["start"]).abs()

    #     counts_df["mean_coverage"] = (self.read_length * counts_df["counts"]) / counts_df["bait_length"]

    #     overall_coverage = sum(counts_df["mean_coverage"]) / len(counts_df)

    #     keep = self.desired_mean_coverage / overall_coverage

    #     return keep

    def ds(self):
        bam_name = os.path.basename(self.alignment_file)
        downsampled_name = f"downsampled_{bam_name}"
        downsampled_path = os.path.join(self.bam_dir, downsampled_name)

        logging.info(f"Downsampling the bam: {bam_name}")

        command = f"gatk DownsampleSam -I {self.alignment_file} -O {downsampled_path} -S {self.strategy} -P {self.keep} --REFERENCE_SEQUENCE {self.refseq}"

        try:
            os.system(command)
            logging.info("DownsampleSam - Successfully downsampled the bam")
        except Exception as e:
            logging.error(f"DownsampleSam - An error occurred while downsampling bam: {e}")

        return downsampled_path

    def remove_unpaired_reads(self, downsampled_bam):  ### YOUREHERE.
        logging.info(f"Removing unpaired reads")

        final_bam_name = self.sample_id + "_final.bam"

        final_bam = os.path.join(self.bam_dir, final_bam_name)

        command = f"samtools view -f 0x2 {downsampled_bam} -b -o {final_bam}"

        try:
            os.system(command)
            logging.info("samtools - Successfully removed unpaired reads")
        except Exception as e:
            logging.error(f"samtools - An error occurred while removing unpaired reads: {e}")

        return final_bam

    def bamtofastq(self, final_bam):
        logging.info(f"Creating fastq files from bam file")

        r1 = os.path.join(self.bam_dir, f"{self.sample_id}_DS_R1_001.fastq.gz")

        r2 = os.path.join(self.bam_dir, f"{self.sample_id}_DS_R2_001.fastq.gz")

        command = f"gatk SamToFastq -I {final_bam} -F {r1} -F2 {r2}"

        try:
            os.system(command)
            logging.info("SamToFastq - Successfully created the downsampled fastq files")
        except Exception as e:
            logging.error(f"SamToFastq - An error occurred while downsampling bam: {e}")


@timer
def downsample(
    sample_id: str,
    r1: str,
    r2: str,
    out_path: str,
    reference: str,
    *,
    threads: int = 40,
    remove_all_dups: bool = False,
    remove_seq_dups: bool = False,
    use_gatk_md: bool = False,
    strategy: str = "HighAccuracy",
    keep: float = 0.5,
) -> None:
    """
    Pipeline to map, deduplicate, and downsample sequencing reads.

    Parameters
    ----------
    sample_id : str
        Sample ID.
    r1 : str
        Path to R1 fastq file.
    r2 : str
        Path to R2 fastq file.
    out_path : str
        Output path for the results.
    reference : str
        Path to the reference genome.
    threads : int
        Number of threads to use , default is 40.
    remove_all_dups : bool
        Whether to remove all duplicates, default is False.
    remove_seq_dups : bool
        Whether to remove sequencing duplicates, default is False.
    use_gatk_md : bool
        Whether to use Use GATK MarkDuplicatesSpark, default is False.
    strategy : str
        Downsampling strategy, default is 'High Accuracy'
    keep : float
        Ratio of the reads to keep [0-1], default is 0.5.
    """
    mapit = Mapper(
        reference_fasta=reference,
        sample_id=sample_id,
        r1=r1,
        r2=r2,
        output_path=out_path,
        threads=threads,
    )
    sam_file = mapit.map()

    dedup = BAMConverter(
        sam_path=sam_file,
        sample_id=sample_id,
        threads=threads,
        use_mark_duplicates=use_gatk_md,
        remove_all_duplicates=remove_all_dups,
        remove_sequencing_duplicates=remove_seq_dups,
    )
    bam_file = dedup.convert_markdedup()

    downsampler = Downsampler(
        alignment_file=bam_file,
        sample_id=sample_id,
        strategy=strategy,
        refseq=reference,
        keep=keep,
    )

    downsampled_bam = downsampler.ds()
    final_bam = downsampler.remove_unpaired_reads(downsampled_bam)
    downsampler.bamtofastq(final_bam)
    return


""" def _old_downsample():
    parser = argparse.ArgumentParser(description="Pipeline to map, deduplicate, and downsample sequencing reads.")
    parser.add_argument("--sample_id", type=str, required=True, help="Sample ID")
    parser.add_argument("--r1", required=True, help="Path to R1 fastq file")
    parser.add_argument("--r2", required=True, help="Path to R2 fastq file")
    parser.add_argument("--out_path", required=True, help="Output path for the results")
    parser.add_argument("--threads", type=int, default=40, help="Number of threads to use")
    parser.add_argument("--remove_all_dups", action="store_true", help="Remove all duplicates")
    parser.add_argument("--remove_seq_dups", action="store_true", help="Remove sequencing duplicates")
    parser.add_argument("--use_gatk_md", action="store_true", help="Use GATK MarkDuplicatesSpark")
    parser.add_argument("--strategy", default="HighAccuracy", help="Downsampling strategy")
    parser.add_argument("--keep", type=float, default=0.5, help="How much read to keep? Give a ratio")
    parser.add_argument("--reference", required=True, help="Path to the reference genome")
    # parser.add_argument('--bed', required=True, help='Path to the exome bait BED file')
    # parser.add_argument('--qc_threshold', type=int, default=20, help='Quality threshold for read counting')
    # parser.add_argument('--read_length', type=int, default=150, help='Read length')
    # parser.add_argument('--desired_mean_coverage', type=int, default=100, help='Desired mean coverage after downsampling')

    args = parser.parse_args()

    mapit = Mapper(
        reference_fasta=args.reference,
        sample_id=args.sample_id,
        r1=args.r1,
        r2=args.r2,
        output_path=args.out_path,
        threads=args.threads,
    )
    sam_file = mapit.map()

    dedup = BAMConverter(
        sam_path=sam_file,
        sample_id=args.sample_id,
        threads=args.threads,
        use_mark_duplicates=args.use_gatk_md,
        remove_all_duplicates=args.remove_all_dups,
        remove_sequencing_duplicates=args.remove_seq_dups,
    )
    bam_file = dedup.convert_markdedup()

    # coverage = CoverageAnalysis(
    #     reference=args.reference,
    #     bam = bam_file,
    #     bed = args.bed,
    #     output_dir=args.out_path,
    #     max_worker= args.threads
    # )
    # coverage.run_mosdepth()
    # coverage.calculate_coverages()
    # coverage.OnTargetContigs()

    downsampler = Downsampler(
        alignment_file=bam_file,
        sample_id=args.sample_id,
        strategy=args.strategy,
        refseq=args.reference,
        keep=args.keep,
        # exome_bait=args.bed,
        # quality_threshold=args.qc_threshold,
        # read_length=args.read_length,
        # desired_mean_coverage=args.desired_mean_coverage
    )
    # counts_file = downsampler.get_bam_counts()
    # keep = downsampler.calculate_keep(counts_file)
    downsampled_bam = downsampler.ds()
    final_bam = downsampler.remove_unpaired_reads(downsampled_bam)
    downsampler.bamtofastq(final_bam)
    return """

if __name__ == "__main__":
    pass



import sys
from pathlib import Path
from subprocess import run

# find the operating system
if sys.platform.startswith("linux"):
    ops = "Linux"
elif sys.platform.startswith("win"):
    ops = "Windows"
elif sys.platform.startswith("darwin"):
    ops = "MacOS"
else:
    ops = "Unknown OS"
# PATHS
home = Path().home().expanduser()
zsh = home / ".zshenv"
bsh = home / ".bash_completion"
# autocomplete text for bashrc or zshrc
text = "\n# BIPackage Autocomplete\n"
text += 'eval "$(register-python-argcomplete bip)"\n'
bshfile = str(Path.home() / ".bashrc")
# bshfile = "deneme.sh"


def _tab_autocomplete_mac():
    """Tab autocomplete for macOS."""
    command = ""
    command += "autoload -U bashcompinit\n"
    command += "bashcompinit\n"
    command += 'eval "$(register-python-argcomplete bip)"\n'
    command += 'eval "$(register-python-argcomplete bipipe)\n'
    # run([command])


def _check_bashrc():
    with open(bshfile) as b:
        bashrc = b.read()

    return text in bashrc


def _tab_autocomplete_linux():
    """Tab autocomplete for Linux."""
    if not (zsh.exists() and bsh.exists()):
        activate = "activate-global-python-argcomplete"
        run([activate])

    if not _check_bashrc():
        with open(bshfile, "a") as f:
            f.write(text)


def _tab_autocomplete():
    if ops == "Linux":
        _tab_autocomplete_linux()
    elif ops == "MacOS":
        _tab_autocomplete_mac()
    else:
        pass


if __name__ == "__main__":
    _tab_autocomplete()



import argparse
import json
import os
from multiprocessing import Pool


class GCNV:
    def __init__(
        self,
        input_json,
        output_dir,
        exome_loc,
        genome_fasta,
        ref_dict,
        contig_pp,
        class_coherence_length,
        cnv_coherence_length,
        min_contig_length,
        p_active,
        p_alt,
        interval_psi_scale,
        sample_psi_scale,
    ):
        self.input_json = input_json
        self.output_dir = output_dir
        self.exome_loc = exome_loc
        self.genome_fasta = genome_fasta
        self.ref_dict = ref_dict
        self.contig_pp = contig_pp
        self.class_coherence_length = class_coherence_length
        self.cnv_coherence_length = cnv_coherence_length
        self.min_contig_length = min_contig_length
        self.p_active = p_active
        self.p_alt = p_alt
        self.interval_psi_scale = interval_psi_scale
        self.sample_psi_scale = sample_psi_scale
        self.input_string = None
        self.filtered_il = None
        self.interval_list = None
        self.ann_interval_list = None
        self.bam_list = None

    def preprocess_intervals(self):
        gatkt = "PreprocessIntervals"
        imr = "OVERLAPPING_ONLY"

        self.interval_list = os.path.join(self.output_dir, "hg38_exome.preprocessed.interval_list")
        command = (
            f"gatk {gatkt} -R {self.genome_fasta} -L {self.exome_loc} --bin-length 0 -imr {imr} -O {self.interval_list}"
        )
        os.system(command)

    def annotate_intervals(self):
        gatkt = "AnnotateIntervals"
        imr = "OVERLAPPING_ONLY"

        self.ann_interval_list = os.path.join(self.output_dir, "hg38_exome.annotated.tsv")
        command = f"gatk {gatkt} -L {self.interval_list} -R {self.genome_fasta} -imr {imr} -O {self.ann_interval_list}"
        os.system(command)

    def run_gatk_collect_rc(self, args):
        bam_path, interval_list, genome_fasta = args
        bam_name = os.path.basename(bam_path)
        sample_name = os.path.splitext(bam_name)[0]
        output_name = sample_name + "_counts.tsv"
        gatkt = "CollectReadCounts"
        imr = "OVERLAPPING_ONLY"

        command = f"gatk {gatkt} -L {interval_list} -R {genome_fasta} -imr {imr} -I {bam_path} -O {os.path.join(self.output_dir, output_name)}"
        os.system(command)

    def get_bam_list(self):
        self.bam_list = []

        with open(self.input_json, "r") as f:
            data = json.load(f)

        self.bam_list = data.get("alignment_files", [])

        print(self.bam_list)

    def count_reads(self):
        arguments = [(bam_path, self.interval_list, self.genome_fasta) for bam_path in self.bam_list]

        try:
            with Pool(len(self.bam_list)) as pool:
                pool.map(self.run_gatk_collect_rc, arguments)

        except Exception as e:
            print(f"error: {e}")

    def get_input_string(self):
        input_string = ""

        self.clist = [
            os.path.join(self.output_dir, item) for item in os.listdir(self.output_dir) if item.endswith("_counts.tsv")
        ]  # 1

        for path in self.clist:
            input_string += f"-I {path} "

        self.input_string = input_string.strip()

        return self.input_string

    def filterintervals(self):
        self.filtered_il = os.path.join(self.output_dir, "cohort_filtered.interval_list")
        gatkt = "FilterIntervals"
        imr = "OVERLAPPING_ONLY"
        command = f"gatk {gatkt} -L {self.interval_list} --annotated-intervals {self.ann_interval_list} {self.input_string} -imr {imr} -O {self.filtered_il}"
        os.system(command)

    def determinecontig(self):
        gatkt = "DetermineGermlineContigPloidy"
        imr = "OVERLAPPING_ONLY"
        command = f"gatk {gatkt} -L {self.filtered_il} -imr {imr} {self.input_string} --contig-ploidy-priors {self.contig_pp} --output {self.output_dir} --output-prefix ploidy --verbosity DEBUG"
        os.system(command)

    def CNVCaller(self):
        gatkt = "GermlineCNVCaller"
        imr = "OVERLAPPING_ONLY"
        ploidy_calls = os.path.join(self.output_dir, "ploidy-calls")
        output = os.path.join(self.output_dir, "cnv-cohort")
        coherence_lengths = (
            f"--class-coherence-length {self.class_coherence_length} --cnv-coherence-length {self.cnv_coherence_length}"
        )
        p_scores = f"--p-active {self.p_active} --p-alt {self.p_alt}"
        psi_scores = f"--interval-psi-scale {self.interval_psi_scale} --sample-psi-scale {self.sample_psi_scale}"
        command = f"gatk {gatkt} --run-mode COHORT -L {self.filtered_il} -imr {imr} {self.input_string} --contig-ploidy-calls {ploidy_calls} --annotated-intervals {self.ann_interval_list} --output {output} --output-prefix cnv-cohort --verbosity DEBUG {coherence_lengths} {p_scores} {psi_scores}"
        print(command)
        os.system(command)

    def postprocess_sample(self):
        gatkt = "PostprocessGermlineCNVCalls"
        imr = "OVERLAPPING_ONLY"
        ploidy_calls = os.path.join(self.output_dir, "ploidy-calls")
        model = os.path.join(self.output_dir, "cnv-cohort", "cnv-cohort-model")
        calls = os.path.join(self.output_dir, "cnv-cohort", "cnv-cohort-calls")

        for i in range(len(self.clist)):
            with open(os.path.join(calls, f"SAMPLE_{i}", "sample_name.txt")) as file:
                name = file.read()
                name = name.strip()

            interval_output = os.path.join(self.output_dir, f"{name}_intervals.vcf.gz")
            segment_output = os.path.join(self.output_dir, f"{name}_segments.vcf.gz")
            denoised = os.path.join(self.output_dir, f"{name}_denoised_copy_ratios.tsv")

            command = f"gatk {gatkt} -imr {imr} --model-shard-path {model} --calls-shard-path {calls} --allosomal-contig chrX --allosomal-contig chrY --contig-ploidy-calls {ploidy_calls} --sample-index {i} --output-genotyped-intervals {interval_output} --output-genotyped-segments {segment_output} --output-denoised-copy-ratios {denoised} --sequence-dictionary {self.ref_dict}"
            print(command)
            os.system(command)

    def panel_of_normals(self):
        gatkt = "CreateReadCountPanelOfNormals"
        output = os.path.join(self.output_dir, "cnvpon.hdf5")
        opts = "--minimum-interval-median-percentile 5.0 --spark-master local[40]"

        command = f"gatk {gatkt} {self.input_string} {opts} -O {output}"

        os.system(command)

        return output

    def stardardize_denoise(self, pon):
        gatkt = "DenoiseReadCounts"
        counts_list = [item for item in os.listdir(self.output_dir) if item.endswith("_counts.tsv")]

        cr_list = []
        for item in counts_list:
            count_path = os.path.join(self.output_dir, item)
            item_prefix = item.split("_counts.tsv")[0]
            sdcr_output = os.path.join(self.output_dir, f"{item_prefix}_standardizedCR.tsv")
            dncr_output = os.path.join(self.output_dir, f"{item_prefix}_denoisedCR.tsv")
            cr_output = [item_prefix, sdcr_output, dncr_output]

            cpon = f"--count-panel-of-normals {pon}"
            sdcr = f"--standardized-copy-ratios {sdcr_output}"
            dncr = f"--denoised-copy-ratios {dncr_output}"

            command = f"gatk {gatkt} -I {count_path} {cpon} {sdcr} {dncr}"

            cr_list.append(cr_output)

            os.system(command)

        return cr_list

    def plot_pt1(self, copy_ratio_paths):
        gatkt = "PlotDenoisedCopyRatios"
        min_contig = f"--minimum-contig-length {self.min_contig_length}"
        output = f"--output {os.path.join(self.output_dir, 'plots')}"
        seq_dict = f"--sequence-dictionary {self.ref_dict}"

        for item in copy_ratio_paths:
            out_prefix = f"--output-prefix {item[0]}"
            sd_cr = f"--standardized-copy-ratios {item[1]}"
            dn_cr = f"--denoised-copy-ratios {item[2]}"

            command = f"gatk {gatkt} {sd_cr} {dn_cr} {seq_dict} {min_contig} {output} {out_prefix}"

            os.system(command)

    def run_gatk_collect_ac(self, args):
        bam_path, interval_list, genome_fasta = args
        bam_name = os.path.basename(bam_path)
        sample_name = os.path.splitext(bam_name)[0]
        output_name = sample_name + "_allelic_counts.tsv"
        gatkt = "CollectAllelicCounts"
        imr = "OVERLAPPING_ONLY"
        command = f"gatk {gatkt} -L {interval_list} -R {genome_fasta} -imr {imr} -I {bam_path} -O {os.path.join(self.output_dir, output_name)}"
        os.system(command)

    def allelic_counts(self):
        arguments = [(bam_path, self.filtered_il, self.genome_fasta) for bam_path in self.bam_list]

        try:
            with Pool(len(self.bam_list)) as pool:
                pool.map(self.run_gatk_collect_ac, arguments)

        except Exception as e:
            print(f"error: {e}")

    def run_gatk_model_segments(self, args):
        sample_name, sample_path, bam_dir = args
        gatkt = "ModelSegments"
        output_dir = os.path.join(bam_dir, "segmentation_results")

        output_prefix = f"--output-prefix {sample_name}"
        denoised = f"--denoised-copy-ratios {sample_path}_denoisedCR.tsv"
        ac = f"--allelic-counts {sample_path}_allelic_counts.tsv"

        command = f"gatk {gatkt} {denoised} {ac} --output {output_dir} {output_prefix}"
        os.system(command)

    def model_segments(self):
        arguments = [
            (
                os.path.splitext(os.path.basename(bam_path))[0],
                os.path.join(self.output_dir, os.path.splitext(os.path.basename(bam_path))[0]),
                self.output_dir,
            )
            for bam_path in self.bam_list
        ]

        try:
            with Pool(len(self.bam_list)) as pool:
                pool.map(self.run_gatk_model_segments, arguments)

        except Exception as e:
            print(f"error: {e}")

        return os.path.join(self.output_dir, "segmentation_results")

    def copy_ratio_segments(self, segment_folder):
        gatkt = "CallCopyRatioSegments"
        input_list = [item for item in os.listdir(segment_folder) if item.endswith(".cr.seg")]

        for item in input_list:
            input = f"--input {os.path.join(segment_folder, item)}"
            sample_name = item.split(".cr.seg")[0]
            output = f"--output {os.path.join(segment_folder, f'{sample_name}.called.seg')}"

            command = f"gatk {gatkt} {input} {output}"
            os.system(command)

    def plot_segments(self, segment_folder):
        gatkt = "PlotModeledSegments"
        output = os.path.join(self.output_dir, "plots_seg")

        for item in self.bam_list:
            sample_name = os.path.splitext(os.path.basename(item))[0]
            denoised = f"--denoised-copy-ratios {os.path.join(self.output_dir, f'{sample_name}_denoisedCR.tsv')}"
            ac = f"--allelic-counts {os.path.join(segment_folder, f'{sample_name}.hets.tsv')}"
            segments = f"--segments {os.path.join(segment_folder, f'{sample_name}.modelFinal.seg')}"
            seq_dict = f"--sequence-dictionary {self.ref_dict}"
            min_contig_l = f"--minimum-contig-length {self.min_contig_length}"
            output_folder = f"--output {output}"
            output_prefix = f"--output-prefix {sample_name}"

            command = (
                f"gatk {gatkt} {denoised} {ac} {segments} {seq_dict} {min_contig_l} {output_folder} {output_prefix}"
            )
            os.system(command)


def main():
    parser = argparse.ArgumentParser(description="Run GCNV pipeline.")
    parser.add_argument("--input_json", type=str, help="Paths to input bams")
    parser.add_argument("--output_dir", type=str, help="Path to output directory")
    parser.add_argument("--exome_loc", type=str, help="Path to exome BED file")
    parser.add_argument("--genome_fasta", type=str, help="Path to genome FASTA file")
    parser.add_argument("--ref_dict", type=str, help="Path to reference dictionary file")
    parser.add_argument("--contig_pp", type=str, help="Path to contig ploidy prior file")
    parser.add_argument(
        "--class_coherence_length",
        type=str,
        default="10000",
        help="Class coherence length",
    )
    parser.add_argument("--cnv_coherence_length", type=str, default="10000", help="CNV coherence length")
    parser.add_argument(
        "--min_contig_length",
        type=str,
        default="46709983",
        help="Minimum contig length",
    )
    parser.add_argument(
        "--p_active",
        type=str,
        default="0.01",
        help="Prior probability of treating an interval as CNV-active",
    )
    parser.add_argument(
        "--p_alt",
        type=str,
        default="0.000001",
        help="Total prior probability of alternative copy-number states",
    )
    parser.add_argument(
        "--interval_psi_scale",
        type=str,
        default="0.001",
        help="Typical scale of interval-specific unexplained variance",
    )
    parser.add_argument(
        "--sample_psi_scale",
        type=str,
        default="0.0001",
        help="Typical scale of sample-specific correction to the unexplained variance",
    )

    args = parser.parse_args()

    counts = GCNV(
        args.input_json,
        args.output_dir,
        args.exome_loc,
        args.genome_fasta,
        args.ref_dict,
        args.contig_pp,
        args.class_coherence_length,
        args.cnv_coherence_length,
        args.min_contig_length,
        args.p_active,
        args.p_alt,
        args.interval_psi_scale,
        args.sample_psi_scale,
    )

    counts.preprocess_intervals()
    counts.annotate_intervals()
    counts.get_bam_list()
    counts.count_reads()
    counts.get_input_string()
    counts.filterintervals()
    counts.determinecontig()
    counts.CNVCaller()
    counts.postprocess_sample()
    # pon = counts.panel_of_normals()
    # cr_list = counts.stardardize_denoise(pon)
    # counts.plot_pt1(cr_list)
    # counts.allelic_counts()
    # seg_dir = counts.model_segments()
    # counts.copy_ratio_segments(seg_dir)
    # counts.plot_segments(seg_dir)

def gatk_gcnv(
    input_json: str,
    output_dir: str,
    exome_loc: str,
    genome_fasta: str,
    ref_dict: str,
    contig_pp: str,
    class_coherence_length: str="10000",
    cnv_coherence_length: str="10000",
    min_contig_length: str="46709983",
    p_active: str="0.01",
    p_alt: str="0.000001",
    interval_psi_scale: str="0.001",
    sample_psi_scale: str="0.0001",
) -> None:
    """
    GATK GCNV pipeline.
    
    Parameters
    ----------
    input_json : str
        Path to JSON file containing input BAM files.
    output_dir : str
        Directory where output files will be saved.
    exome_loc : str
        Path to BED file defining the exome regions.
    genome_fasta : str
        Path to the reference genome FASTA file.
    ref_dict : str
        Path to the reference dictionary file.
    contig_pp : str
        Path to the contig ploidy prior file.
    class_coherence_length : str, optional
        Class coherence length, by default "10000".
    cnv_coherence_length : str, optional
        CNV coherence length, by default "10000".
    min_contig_length : str, optional
        Minimum contig length, by default "46709983".
    p_active : str, optional
        Prior probability of treating an interval as CNV-active, by default "0.01".
    p_alt : str, optional
        Total prior probability of alternative copy-number states, by default "0.000001".
    interval_psi_scale : str, optional
        Scale of interval-specific unexplained variance, by default "0.001".
    sample_psi_scale : str, optional
        Scale of sample-specific correction to unexplained variance, by default "0.0001".
    """
    counts = GCNV(
        input_json,
        output_dir,
        exome_loc,
        genome_fasta,
        ref_dict,
        contig_pp,
        class_coherence_length,
        cnv_coherence_length,
        min_contig_length,
        p_active,
        p_alt,
        interval_psi_scale,
        sample_psi_scale,
    )

    counts.preprocess_intervals()
    counts.annotate_intervals()
    counts.get_bam_list()
    counts.count_reads()
    counts.get_input_string()
    counts.filterintervals()
    counts.determinecontig()
    counts.CNVCaller()
    counts.postprocess_sample()
    # pon = counts.panel_of_normals()
    # cr_list = counts.stardardize_denoise(pon)
    # counts.plot_pt1(cr_list)
    # counts.allelic_counts()
    # seg_dir = counts.model_segments()
    # counts.copy_ratio_segments(seg_dir)
    # counts.plot_segments(seg_dir)
    return None

if __name__ == "__main__":
    main()



import argparse
import gzip
import zlib


# subcommand
def check_gzip_validity(file_path: str) -> None:
    """
    Check a compressed file validity, prints results to stdout.

    Parameters
    ----------
    file_path
        Path to the compressed file.

    Returns
    -------
    None
    """
    try:
        with gzip.open(file_path, "rb") as file:
            file.read()
        print("File read successfully.")
    except gzip.BadGzipFile:
        print("The file is truncated or invalid.")
    except zlib.error:
        print("The file is truncated or invalid.")
    except IOError:
        print("Error reading file")

    return


"""def _test_check_gzip_validity()
    parser = argparse.ArgumentParser(description="It checks a single compressed file in a specified directory")
    parser.add_argument("file_path", help="Location of the file")

    args = parser.parse_args()

    check_gzip_validity(args.file_path)
    return"""


if __name__ == "__main__":
    pass



import argparse
import ftplib
import os
import subprocess

from bipackage.util.utilities import timer


'''Gets a bcl file, demultiplex it and send relevant fastqs to a remote server'''

class NIPTBcl2Fastq:
    def __init__(
        self,
        output_folder,
        num_readers,
        num_writers,
        num_processors,
        compression_level,
        fastq_file_names,
        nipt_part,
        nipt_folders,
        source_path,
        bcl_path,
    ):
        self.output_folder = output_folder
        self.num_readers = num_readers
        self.num_writers = num_writers
        self.num_processors = num_processors
        self.compression_level = compression_level
        self.nipt_part = nipt_part
        self.fastq_file_names = fastq_file_names
        self.bclfolders = nipt_folders
        self.source_path = source_path
        self.bcl_path = bcl_path
        self.fastq_folder_paths = []
        self.fastq_file_paths = []
        self.copybclfromnas()
        self.create_folders()
        self.bcl_to_fastq()
        self.take_sample_paths()
        # self.upload_file_secure()

    def copybclfromnas(self):
        print("Copying bcl files to server")

        if self.bclfolders:
            source_paths = " ".join([f"{self.source_path}/{folder}" for folder in self.bclfolders])
            command = f"cp -r {source_paths} {self.bcl_path}"

            subprocess.run(command, shell=True)
        else:
            print("No BCL folders found to copy.")

    def create_folders(self):
        print("Creating fastq folders")
        for folder in self.bclfolders:
            fastq_folder = os.path.join(self.output_folder, folder)
            os.makedirs(fastq_folder, exist_ok=True)
            self.fastq_folder_paths.append(fastq_folder)
        print("Fastq folders created")

    def bcl_to_fastq(self):
        print("Starting bcl2fastq")

        self.fastq_folder_paths = []

        for bcl_folder in self.bclfolders:
            bcl_abs_path = os.path.join(self.bcl_path, bcl_folder)
            fastq_abs_path = os.path.join(self.output_folder, bcl_folder)
            samplesheet_path = os.path.join(bcl_abs_path, "SampleSheet.csv")

            self.fastq_folder_paths.append(fastq_abs_path)

            run_command = (
                f"bcl2fastq -R {bcl_abs_path} -o {fastq_abs_path} "
                f"--sample-sheet {samplesheet_path} -r {self.num_readers} "
                f"-w {self.num_writers} -p {self.num_processors} "
                f"--fastq-compression-level {self.compression_level} "
                "--no-lane-splitting"
            )

            subprocess.run(run_command, shell=True)

    def take_sample_paths(self):
        print("Taking sample paths")

        all_folder_paths = [os.path.join(self.output_folder, folder) for folder in self.bclfolders]
        print(all_folder_paths)
        self.fastq_file_paths = []

        for folder_path in all_folder_paths:
            for root, _, files in os.walk(folder_path):
                for file in files:
                    for file_name in self.fastq_file_names:
                        if file.startswith(file_name):
                            self.fastq_file_paths.append(os.path.join(root, file))

        print("Took sample paths")

    def upload_file_secure(self, ftp_server, ftp_user, ftp_password):
        print("Started uploading the files")
        try:
            ftps = ftplib.FTP_TLS()
            ftps.connect(ftp_server, 21)
            ftps.login(user=ftp_user, passwd=ftp_password)
            ftps.prot_p()
            ftps.set_pasv(True)
            counter = 0
            for file_path in self.fastq_file_paths:
                folder_path = file_path.split("/")[-2]
                fastq_name = os.path.basename(file_path)
                remote_path = os.path.join(f"NIPT-Part{str(self.nipt_part)}", folder_path, fastq_name)
                with open(file_path, "rb") as file:
                    ftps.storbinary(f"STOR {remote_path}", file)
                    print(f"File uploaded successfully: {remote_path}")
                    counter += 1
            print(f"In total {counter} Fastq files uploaded to the server")
            ftps.quit()
        except ftplib.all_errors as e:
            print(f"FTP error: {e}")

        return


@timer
def nipt_bcl2fastq(
    *,
    nipt_folders: str | list[str],
    nipt_part: str,
    fastq_names: str | list[str],
    output_folder: str,
    source_path: str,
    bcl_path: str,
    num_readers: int = 10,
    num_writers: int = 10,
    num_processors: int = 40,
    compression_level: int = 8,
) -> None:
    """
    Run bcl2fastq conversion for multiple BCL folders, upload resulting fastqs to the server.

    Parameters
    ----------
    nipt_folders : str | list[str]
        Nipt folders.
    nipt_part: str
        Part number.
    fastq_names: str | list[str]
        Fastq sample names - For example 24B3043312
    output_folder: str
        Path to the output Fastq folder.
    num_readers: int = 10
        Number of readers.
    num_writers: int = 10
        Number of writers.
    num_processors: int = 40
        Number of processors.
    compression_level: int = 8
        Compression level.
    source_path: str = source_path
        Path to the source folder.
    bcl_path: str = bcl_path
        Path to the bcl folder.

    """
    nipt = NIPTBcl2Fastq(
        output_folder,
        num_readers,
        num_writers,
        num_processors,
        compression_level,
        fastq_names,
        nipt_part,
        nipt_folders,
        source_path=source_path,
        bcl_path=bcl_path,
    )
    return nipt


if __name__ == "__main__":
    pass



from bipackage.src.ittools.ismount import check_reconnect, is_mounted, mount_server
from bipackage.src.ittools.md5sumchecker_multi import md5sumchecker
from bipackage.src.ittools.truncatedfchecker_single import check_gzip_validity

__all__ = [
    "check_gzip_validity",
    "check_reconnect",
    "is_mounted",
    "md5sumchecker",
    "mount_server",
]



import argparse
import multiprocessing
import os
import subprocess

from bipackage.util.utilities import timer


'''Checks md5sums of files and validates if they match'''

class MD5SumChecker:
    def __init__(self, input_directory, file_extension, num_processes):
        self.input_directory = input_directory
        self.file_extension = file_extension
        self.num_processes = num_processes
        self.checker()

    def calculate_md5sum(self, file_path):
        try:
            result = subprocess.run(
                ["md5sum", file_path],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                check=True,
            )

            md5sum_value = result.stdout.split()[0]

            return md5sum_value
        except subprocess.CalledProcessError as e:
            print(f"Error: {e}")
            return None

    def check_file(self, root, filename):
        md5sum_name = filename + ".md5sum"
        with open(os.path.join(root, md5sum_name), "r") as file:
            md5sum_file = file.read()
            md5sum_value = md5sum_file.split()[0]

        md5sum_path = os.path.join(root, md5sum_name)
        f_path = os.path.join(root, filename)

        calcmd5sum = self.calculate_md5sum(f_path)

        if calcmd5sum is not None and md5sum_value == calcmd5sum:
            print(f"For {f_path} {md5sum_path}, md5sums are same")
        else:
            print(f"For {f_path} {md5sum_path}, md5sums are different")

        return

    def checker(self):
        pool = multiprocessing.Pool(processes=self.num_processes)
        for root, dirs, files in os.walk(self.input_directory):
            for filename in files:
                if filename.endswith(self.file_extension):
                    pool.apply_async(self.check_file, args=(root, filename))
        pool.close()
        pool.join()

        return


# subcommand
@timer
def md5sumchecker(input_directory: str, file_extension, num_processes: int = 2) -> None:
    """
    MD5Sum Checker.

    Parameters
    ----------
    input_directory : str
        Input directory path.
    file_extension : str
        File extension to search for.
    num_processes : int
        Number of processes to use, default is 2

    Returns
    -------
    MD5SumChecker
    """
    md5_checker = MD5SumChecker(
        input_directory=input_directory, file_extension=file_extension, num_processes=num_processes
    )
    return md5_checker


def main():
    parser = argparse.ArgumentParser(description="MD5Sum Checker")
    parser.add_argument("-i", "--input_directory", required=True, help="Input directory path")
    parser.add_argument("-fe", "--file_extension", required=True, help="File extension to search for")
    parser.add_argument(
        "-p",
        "--num_processes",
        type=int,
        default=2,
        help="Number of processes to use (default: 2)",
    )
    args = parser.parse_args()

    md5_checker = MD5SumChecker(args.input_directory, args.file_extension, args.num_processes)


if __name__ == "__main__":
    pass



import argparse
import logging
import os
import subprocess

import pandas as pd

from bipackage.constants import PARSED_GTF_PATH_GRCh38, WHOLE_GENE_LOCS_PATH_GRCh38
from bipackage.util.utilities import timer


class BedFileGenerator:
    def __init__(
        self,
        parsed_gtf_path,
        whole_gene_locs_path,
        gene_list,
        bed_file_name,
        output_folder,
        whole_gene_list=None,
        cds=True,
    ):
        self.parsed_gtf_path = parsed_gtf_path
        self.whole_gene_locs_path = whole_gene_locs_path
        self.gene_list = gene_list
        self.cds = cds
        self.output_folder = output_folder
        self.bed_file_name = bed_file_name
        self.whole_gene_list = whole_gene_list
        self.bed_file_path = None

    def bed_creator(self):
        logging.info(f"Creating bed file: {self.bed_file_name}")

        gtf = pd.read_csv(self.parsed_gtf_path, sep="\t", low_memory=False)

        gtf_all_feats = gtf[
            [
                "seqname",
                "feature",
                "start",
                "end",
                "gene_name",
                "gene_id",
                "tag",
                "exon_number",
            ]
        ]

        if self.cds:
            gtf_feature = gtf_all_feats[gtf_all_feats["feature"] == "CDS"]
        else:
            gtf_feature = gtf_all_feats[gtf_all_feats["feature"] == "exon"]

        missing_genes = set(self.gene_list) - set(gtf_feature["gene_name"])

        if missing_genes:
            raise ValueError(f"The following gene names are not in the 'gene_name' column: {', '.join(missing_genes)}")

        genes = gtf_feature[gtf_feature["gene_name"].isin(self.gene_list)]

        canonicals = genes[genes["tag"].str.contains("MANE_Select|Ensembl_canonical", na=False)]

        canonicals["seqname"] = "chr" + canonicals["seqname"].astype(str)

        canonicals["seqname"] = canonicals["seqname"].str.strip()

        canonicals["gene_exon"] = canonicals["gene_name"] + "|" + canonicals["exon_number"].astype(str)

        canonicals_filtered = canonicals[["seqname", "start", "end", "gene_exon"]]

        if self.whole_gene_list:
            wglocs = pd.read_csv(self.whole_gene_locs_path, sep="\t", low_memory=False, header=None)
            wglocs.columns = ["seqname", "start", "end", "gene_exon"]
            wgenes = wglocs[wglocs["gene_exon"].isin(self.whole_gene_list)]
            whole_gene_count = wgenes["gene_exon"].nunique()

            missing_whole_genes = set(self.whole_gene_list) - set(wgenes["gene_exon"])

            if missing_whole_genes:
                raise ValueError(
                    f"The following whole gene names are not in the 'gene_name' column: {', '.join(missing_whole_genes)}"
                )

            final_bed = pd.concat([canonicals_filtered, wgenes], ignore_index=True)

        else:
            final_bed = canonicals_filtered
            whole_gene_count = 0

        self.bed_file_path = os.path.join(self.output_folder, f"{self.bed_file_name}.bed")

        final_bed.to_csv(self.bed_file_path, sep="\t", header=False, index=False)

        unique_gene_count = canonicals["gene_name"].nunique()

        logging.info(
            f"Created bed file: {self.bed_file_name} with {unique_gene_count} genes and with {whole_gene_count} whole genes"
        )

    def sort_bed(self):
        sorted_bed_file_path = os.path.join(self.output_folder, f"sorted_{self.bed_file_name}.bed")
        command = f"bedtools sort -i {self.bed_file_path} > {sorted_bed_file_path}"

        try:
            subprocess.run(command, shell=True, check=True)
            logging.info(f"Successfully sorted the BED file. Output saved to {sorted_bed_file_path}")
        except subprocess.CalledProcessError as e:
            logging.error(f"An error occurred while sorting the BED file: {e}")
        except Exception as e:
            logging.error(f"An unexpected error occurred: {e}")


@timer
def bedfilegenerator(
    *,
    gene_list: str | list[str],
    bed_file_name: str,
    output_folder: str,
    whole_gene_list: str | list[str] | None,
    ref_folder: str | None = None,
    parsed_gtf_path: str | None = None,
    whole_gene_locs_path: str | None = None,
    cds: bool = False,
):
    """
    The main command for the CLI entry point. Generate and sort BED files from a parsed GTF file.

    Parameters
    ----------
    gene_list: str | list[str]
        List of gene names to include.
    bed_file_name: str
        Name of the output BED file.
    output_folder: str
        Folder to save the output BED file.
    whole_gene_list: str | list[str]
        List of whole gene names to include.
    ref_folder : str | None
        Path to the folder for reference files. e.g "fasta" or "gtf".
        Will be concatenated with the default paths.
    parsed_gtf_path : str
        Path to Parsed GTF file. e.g "Homo_sapiens.GRCh38.112.parsed.gtf"
    whole_gene_locs_path : str
        Path to whole gene locs file. e.g "Homo_sapiens.GRCh38.112.gtf.gene_names.sorted.cleaned.bed"
    cds: bool = False,
        If True, use CDS features; otherwise use exon features.
    """
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")

    if ref_folder is not None:
        if parsed_gtf_path is None:
            parsed_gtf_path = os.path.join(ref_folder, PARSED_GTF_PATH_GRCh38)
        if whole_gene_locs_path in None:
            whole_gene_locs_path = os.path.join(ref_folder, WHOLE_GENE_LOCS_PATH_GRCh38)

    # guard clauses
    if isinstance(gene_list, str):
        gene_list = list[str]
    if isinstance(whole_gene_list, str):
        whole_gene_list = list[whole_gene_list]

    # create the instance
    bed_generator = BedFileGenerator(
        parsed_gtf_path=parsed_gtf_path,
        whole_gene_locs_path=whole_gene_locs_path,
        gene_list=gene_list,
        bed_file_name=bed_file_name,
        output_folder=output_folder,
        whole_gene_list=whole_gene_list,
        cds=cds,
    )
    # perform the operations
    bed_generator.bed_creator()
    bed_generator.sort_bed()

    return


if __name__ == "__main__":
    pass



from bipackage.src.fastqtools.downsample import downsample
from bipackage.src.fastqtools.fastq_read_counter import fastq_read_counter
from bipackage.src.fastqtools.fastqvalidate import fastqvalidate
from bipackage.src.fastqtools.merge_it import merge_it
from bipackage.src.fastqtools.remove_undetermined_fastq import remove_undetermined_fastq
from bipackage.src.fastqtools.undetermined_demultiplexer import undetermined_demultiplexer

__all__ = [
    "downsample",
    "fastq_read_counter",
    "fastqvalidate",
    "merge_it",
    "remove_undetermined_fastq",
    "undetermined_demultiplexer",
]



import os

from bipackage.util.utilities import timer


def check_files(directory: str):
    for filename in os.listdir(directory):
        if filename.endswith("fastq.gz") or filename.endswith("fq.gz"):
            if os.path.isfile(os.path.join(directory, filename)):
                print("Checking file:", os.path.join(directory, filename))
                print("Performing validation for file:", filename)
                command = f"fastQValidator --file {os.path.join(directory, filename)} --noeof"
                print("Command:", command)
                os.system(command)
        else:
            pass


@timer
def fastqvalidate(directory: str) -> None:
    """
    Validate fastq files in a given directory using`fastQValidator`.

    Parameters
    ----------
    directory
        Directory to perform the validation of fastq files.

    Returns
    -------
    None
    """
    check_files(directory=directory)
    return


if __name__ == "__main__":
    pass



from bipackage.util.aliaser import aliaser


def _read_genes(filename: str = "genes.tsv"):
    """
    Reads a gene list file, where each gene name seperated by a new line and return the gene list.

    Example gene list file (genes.tsv):
    --------
    BRWD3
    POU3F4
    CHM
    GLA
    PLP1
    PRPS1
    ...
    """
    with open(filename) as f:
        content = f.read().rstrip("\n")

    genes = content.split("\n")
    return genes


def _match_genes(
    filename: str,
    genes: list[str],
    *,
    gene_column_index: int = -1,
) -> list[str]:
    """Matches genes list with genes within bedfile."""
    genes_exist = []
    with open(filename) as file:
        for line in file:
            gene_name = line.split("\t")[gene_column_index].rstrip("\n")
            if gene_name in genes:
                if gene_name not in genes_exist:
                    genes_exist.append(gene_name)

    return genes_exist


def panelgenequery(
    bedfile: str,
    genes_list_file: str,
    *,
    gene_column_index: int = -1,
) -> None:
    """
    Creates a modified version of gene_list_file the second column confirms the presence of a gene.

    bedfile: str
        Path to the intersected bedfile where the last column should be the gene name.
        If the gene name column is different, use the `gene_column_index` argument.
    genes_list_file: str
        Path to the gene list file. Each gene name must be seperated by a new line.
    gene_column_index: int
        Index of the column in the bedfile that contains the genes.
        Defaults is -1 (the last column).

    Returns
    -------
    None
    """
    genes = _read_genes(genes_list_file)
    genes_exist = _match_genes(bedfile, genes=genes, gene_column_index=gene_column_index)
    genes_exist = sorted(genes_exist)
    genes_not_exist = [gene for gene in genes if gene not in genes_exist]

    try:
        alias_present = {}
        for gene in genes_not_exist[:]:
            print(f"{gene} could not be found, checking aliases.")
            aliases = aliaser(gene_symbol=gene)
            if aliases != []:
                print(f"aliases for {gene} : {aliases} ")
                matched_aliases = _match_genes(bedfile, genes=aliases, gene_column_index=gene_column_index)
                if matched_aliases != []:
                    print(f"matched aliases : {matched_aliases}")
                    genes_not_exist.remove(gene)
                    genes_exist.append(gene)
                    alias_present[gene] = matched_aliases[0]
    except Exception as e:
        print("Error occurred while checking aliases. Skipping aliases...")
        print(f"Exception: {e}")


    genes_exist = sorted(genes_exist)


    extension = genes_list_file.split(".")[-1]
    exists_path = genes_list_file.replace(extension, f"exist.{extension}")
    with open(exists_path, "w+") as ep:
        for gene in genes_not_exist:
            ep.write(f"{gene}\tNO\n")
        for gene in genes_exist:
            if gene in alias_present:
                ep.write(f"{gene}[{alias_present.get(gene)}]\tYES\n")
            else:
                ep.write(f"{gene}\tYES\n")
    return


